from typing import ClassVar, Dict, Optional, List

from attrs import define, field

from resoto_plugin_aws.resource.base import AwsResource, AwsApiSpec, GraphBuilder
from resotolib.json_bender import Bender, S, Bend, ForallBend, K, F
from resoto_plugin_aws.aws_client import AwsClient
from resoto_plugin_aws.utils import ToDict
from typing import Tuple, Type
from datetime import datetime
from resotolib.types import Json
from resoto_plugin_aws.resource.ec2 import AwsEc2Vpc, AwsEc2SecurityGroup, AwsEc2Subnet
from resoto_plugin_aws.resource.iam import AwsIamRole


@define(eq=False, slots=False)
class AwsRedshiftNetworkInterface:
    kind: ClassVar[str] = "aws_redshift_network_interface"
    mapping: ClassVar[Dict[str, Bender]] = {
        "network_interface_id": S("NetworkInterfaceId"),
        "subnet_id": S("SubnetId"),
        "private_ip_address": S("PrivateIpAddress"),
        "availability_zone": S("AvailabilityZone"),
    }
    network_interface_id: Optional[str] = field(default=None)
    subnet_id: Optional[str] = field(default=None)
    private_ip_address: Optional[str] = field(default=None)
    availability_zone: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsRedshiftVpcEndpoint:
    kind: ClassVar[str] = "aws_redshift_vpc_endpoint"
    mapping: ClassVar[Dict[str, Bender]] = {
        "vpc_endpoint_id": S("VpcEndpointId"),
        "vpc_id": S("VpcId"),
        "network_interfaces": S("NetworkInterfaces", default=[]) >> ForallBend(AwsRedshiftNetworkInterface.mapping),
    }
    vpc_endpoint_id: Optional[str] = field(default=None)
    vpc_id: Optional[str] = field(default=None)
    network_interfaces: List[AwsRedshiftNetworkInterface] = field(factory=list)


@define(eq=False, slots=False)
class AwsRedshiftEndpoint:
    kind: ClassVar[str] = "aws_redshift_endpoint"
    mapping: ClassVar[Dict[str, Bender]] = {
        "address": S("Address"),
        "port": S("Port"),
        "vpc_endpoints": S("VpcEndpoints", default=[]) >> ForallBend(AwsRedshiftVpcEndpoint.mapping),
    }
    address: Optional[str] = field(default=None)
    port: Optional[int] = field(default=None)
    vpc_endpoints: List[AwsRedshiftVpcEndpoint] = field(factory=list)


@define(eq=False, slots=False)
class AwsRedshiftClusterSecurityGroupMembership:
    kind: ClassVar[str] = "aws_redshift_cluster_security_group_membership"
    mapping: ClassVar[Dict[str, Bender]] = {
        "cluster_security_group_name": S("ClusterSecurityGroupName"),
        "status": S("Status"),
    }
    cluster_security_group_name: Optional[str] = field(default=None)
    status: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsRedshiftVpcSecurityGroupMembership:
    kind: ClassVar[str] = "aws_redshift_vpc_security_group_membership"
    mapping: ClassVar[Dict[str, Bender]] = {"vpc_security_group_id": S("VpcSecurityGroupId"), "status": S("Status")}
    vpc_security_group_id: Optional[str] = field(default=None)
    status: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsRedshiftClusterParameterStatus:
    kind: ClassVar[str] = "aws_redshift_cluster_parameter_status"
    mapping: ClassVar[Dict[str, Bender]] = {
        "parameter_name": S("ParameterName"),
        "parameter_apply_status": S("ParameterApplyStatus"),
        "parameter_apply_error_description": S("ParameterApplyErrorDescription"),
    }
    parameter_name: Optional[str] = field(default=None)
    parameter_apply_status: Optional[str] = field(default=None)
    parameter_apply_error_description: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsRedshiftClusterParameterGroupStatus:
    kind: ClassVar[str] = "aws_redshift_cluster_parameter_group_status"
    mapping: ClassVar[Dict[str, Bender]] = {
        "parameter_group_name": S("ParameterGroupName"),
        "parameter_apply_status": S("ParameterApplyStatus"),
        "cluster_parameter_status_list": S("ClusterParameterStatusList", default=[])
        >> ForallBend(AwsRedshiftClusterParameterStatus.mapping),
    }
    parameter_group_name: Optional[str] = field(default=None)
    parameter_apply_status: Optional[str] = field(default=None)
    cluster_parameter_status_list: List[AwsRedshiftClusterParameterStatus] = field(factory=list)


@define(eq=False, slots=False)
class AwsRedshiftPendingModifiedValues:
    kind: ClassVar[str] = "aws_redshift_pending_modified_values"
    mapping: ClassVar[Dict[str, Bender]] = {
        "master_user_password": S("MasterUserPassword"),
        "node_type": S("NodeType"),
        "number_of_nodes": S("NumberOfNodes"),
        "cluster_type": S("ClusterType"),
        "cluster_version": S("ClusterVersion"),
        "automated_snapshot_retention_period": S("AutomatedSnapshotRetentionPeriod"),
        "cluster_identifier": S("ClusterIdentifier"),
        "publicly_accessible": S("PubliclyAccessible"),
        "enhanced_vpc_routing": S("EnhancedVpcRouting"),
        "maintenance_track_name": S("MaintenanceTrackName"),
        "encryption_type": S("EncryptionType"),
    }
    master_user_password: Optional[str] = field(default=None)
    node_type: Optional[str] = field(default=None)
    number_of_nodes: Optional[int] = field(default=None)
    cluster_type: Optional[str] = field(default=None)
    cluster_version: Optional[str] = field(default=None)
    automated_snapshot_retention_period: Optional[int] = field(default=None)
    cluster_identifier: Optional[str] = field(default=None)
    publicly_accessible: Optional[bool] = field(default=None)
    enhanced_vpc_routing: Optional[bool] = field(default=None)
    maintenance_track_name: Optional[str] = field(default=None)
    encryption_type: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsRedshiftRestoreStatus:
    kind: ClassVar[str] = "aws_redshift_restore_status"
    mapping: ClassVar[Dict[str, Bender]] = {
        "status": S("Status"),
        "current_restore_rate_in_mega_bytes_per_second": S("CurrentRestoreRateInMegaBytesPerSecond"),
        "snapshot_size_in_mega_bytes": S("SnapshotSizeInMegaBytes"),
        "progress_in_mega_bytes": S("ProgressInMegaBytes"),
        "elapsed_time_in_seconds": S("ElapsedTimeInSeconds"),
        "estimated_time_to_completion_in_seconds": S("EstimatedTimeToCompletionInSeconds"),
    }
    status: Optional[str] = field(default=None)
    current_restore_rate_in_mega_bytes_per_second: Optional[float] = field(default=None)
    snapshot_size_in_mega_bytes: Optional[int] = field(default=None)
    progress_in_mega_bytes: Optional[int] = field(default=None)
    elapsed_time_in_seconds: Optional[int] = field(default=None)
    estimated_time_to_completion_in_seconds: Optional[int] = field(default=None)


@define(eq=False, slots=False)
class AwsRedshiftDataTransferProgress:
    kind: ClassVar[str] = "aws_redshift_data_transfer_progress"
    mapping: ClassVar[Dict[str, Bender]] = {
        "status": S("Status"),
        "current_rate_in_mega_bytes_per_second": S("CurrentRateInMegaBytesPerSecond"),
        "total_data_in_mega_bytes": S("TotalDataInMegaBytes"),
        "data_transferred_in_mega_bytes": S("DataTransferredInMegaBytes"),
        "estimated_time_to_completion_in_seconds": S("EstimatedTimeToCompletionInSeconds"),
        "elapsed_time_in_seconds": S("ElapsedTimeInSeconds"),
    }
    status: Optional[str] = field(default=None)
    current_rate_in_mega_bytes_per_second: Optional[float] = field(default=None)
    total_data_in_mega_bytes: Optional[int] = field(default=None)
    data_transferred_in_mega_bytes: Optional[int] = field(default=None)
    estimated_time_to_completion_in_seconds: Optional[int] = field(default=None)
    elapsed_time_in_seconds: Optional[int] = field(default=None)


@define(eq=False, slots=False)
class AwsRedshiftHsmStatus:
    kind: ClassVar[str] = "aws_redshift_hsm_status"
    mapping: ClassVar[Dict[str, Bender]] = {
        "hsm_client_certificate_identifier": S("HsmClientCertificateIdentifier"),
        "hsm_configuration_identifier": S("HsmConfigurationIdentifier"),
        "status": S("Status"),
    }
    hsm_client_certificate_identifier: Optional[str] = field(default=None)
    hsm_configuration_identifier: Optional[str] = field(default=None)
    status: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsRedshiftClusterSnapshotCopyStatus:
    kind: ClassVar[str] = "aws_redshift_cluster_snapshot_copy_status"
    mapping: ClassVar[Dict[str, Bender]] = {
        "destination_region": S("DestinationRegion"),
        "retention_period": S("RetentionPeriod"),
        "manual_snapshot_retention_period": S("ManualSnapshotRetentionPeriod"),
        "snapshot_copy_grant_name": S("SnapshotCopyGrantName"),
    }
    destination_region: Optional[str] = field(default=None)
    retention_period: Optional[int] = field(default=None)
    manual_snapshot_retention_period: Optional[int] = field(default=None)
    snapshot_copy_grant_name: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsRedshiftClusterNode:
    kind: ClassVar[str] = "aws_redshift_cluster_node"
    mapping: ClassVar[Dict[str, Bender]] = {
        "node_role": S("NodeRole"),
        "private_ip_address": S("PrivateIPAddress"),
        "public_ip_address": S("PublicIPAddress"),
    }
    node_role: Optional[str] = field(default=None)
    private_ip_address: Optional[str] = field(default=None)
    public_ip_address: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsRedshiftElasticIpStatus:
    kind: ClassVar[str] = "aws_redshift_elastic_ip_status"
    mapping: ClassVar[Dict[str, Bender]] = {"elastic_ip": S("ElasticIp"), "status": S("Status")}
    elastic_ip: Optional[str] = field(default=None)
    status: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsRedshiftClusterIamRole:
    kind: ClassVar[str] = "aws_redshift_cluster_iam_role"
    mapping: ClassVar[Dict[str, Bender]] = {"iam_role_arn": S("IamRoleArn"), "apply_status": S("ApplyStatus")}
    iam_role_arn: Optional[str] = field(default=None)
    apply_status: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsRedshiftDeferredMaintenanceWindow:
    kind: ClassVar[str] = "aws_redshift_deferred_maintenance_window"
    mapping: ClassVar[Dict[str, Bender]] = {
        "defer_maintenance_identifier": S("DeferMaintenanceIdentifier"),
        "defer_maintenance_start_time": S("DeferMaintenanceStartTime"),
        "defer_maintenance_end_time": S("DeferMaintenanceEndTime"),
    }
    defer_maintenance_identifier: Optional[str] = field(default=None)
    defer_maintenance_start_time: Optional[datetime] = field(default=None)
    defer_maintenance_end_time: Optional[datetime] = field(default=None)


@define(eq=False, slots=False)
class AwsRedshiftResizeInfo:
    kind: ClassVar[str] = "aws_redshift_resize_info"
    mapping: ClassVar[Dict[str, Bender]] = {
        "resize_type": S("ResizeType"),
        "allow_cancel_resize": S("AllowCancelResize"),
    }
    resize_type: Optional[str] = field(default=None)
    allow_cancel_resize: Optional[bool] = field(default=None)


@define(eq=False, slots=False)
class AwsRedshiftAquaConfiguration:
    kind: ClassVar[str] = "aws_redshift_aqua_configuration"
    mapping: ClassVar[Dict[str, Bender]] = {
        "aqua_status": S("AquaStatus"),
        "aqua_configuration_status": S("AquaConfigurationStatus"),
    }
    aqua_status: Optional[str] = field(default=None)
    aqua_configuration_status: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsRedshiftReservedNodeExchangeStatus:
    kind: ClassVar[str] = "aws_redshift_reserved_node_exchange_status"
    mapping: ClassVar[Dict[str, Bender]] = {
        "reserved_node_exchange_request_id": S("ReservedNodeExchangeRequestId"),
        "status": S("Status"),
        "request_time": S("RequestTime"),
        "source_reserved_node_id": S("SourceReservedNodeId"),
        "source_reserved_node_type": S("SourceReservedNodeType"),
        "source_reserved_node_count": S("SourceReservedNodeCount"),
        "target_reserved_node_offering_id": S("TargetReservedNodeOfferingId"),
        "target_reserved_node_type": S("TargetReservedNodeType"),
        "target_reserved_node_count": S("TargetReservedNodeCount"),
    }
    reserved_node_exchange_request_id: Optional[str] = field(default=None)
    status: Optional[str] = field(default=None)
    request_time: Optional[datetime] = field(default=None)
    source_reserved_node_id: Optional[str] = field(default=None)
    source_reserved_node_type: Optional[str] = field(default=None)
    source_reserved_node_count: Optional[int] = field(default=None)
    target_reserved_node_offering_id: Optional[str] = field(default=None)
    target_reserved_node_type: Optional[str] = field(default=None)
    target_reserved_node_count: Optional[int] = field(default=None)


# AWS does not give us a cluster arn in the api, and we have to figure it out
# from the cluster network arn and the cluster name.
def _cluster_arn(ns_arn_and_cluster_id: Tuple[str, str]) -> str:
    ns_arn = ns_arn_and_cluster_id[0]
    cluster_id = ns_arn_and_cluster_id[1]
    splitted = ns_arn.split(":")
    return f"arn:aws:redshift:{splitted[3]}:{splitted[4]}:cluster:{cluster_id[1]}"


@define(eq=False, slots=False)
class AwsRedshiftCluster(AwsResource):
    kind: ClassVar[str] = "aws_redshift_cluster"
    api_spec: ClassVar[AwsApiSpec] = AwsApiSpec("redshift", "describe-clusters", "Clusters")

    mapping: ClassVar[Dict[str, Bender]] = {
        "id": S("ClusterIdentifier"),
        "tags": S("Tags", default=[]) >> ToDict(),
        "name": S("ClusterIdentifier"),
        "arn": ((S("ClusterNamespaceArn") >> F(lambda x: (x,))) + (S("ClusterIdentifier") >> F(lambda x: (x,))))
        >> F(lambda cn: _cluster_arn(cn)),
        "ctime": S("ClusterCreateTime"),
        "mtime": K(None),
        "atime": K(None),
        "redshift_node_type": S("NodeType"),
        "redshift_cluster_status": S("ClusterStatus"),
        "redshift_cluster_availability_status": S("ClusterAvailabilityStatus"),
        "redshift_modify_status": S("ModifyStatus"),
        "redshift_master_username": S("MasterUsername"),
        "redshift_db_name": S("DBName"),
        "redshift_endpoint": S("Endpoint") >> Bend(AwsRedshiftEndpoint.mapping),
        "redshift_automated_snapshot_retention_period": S("AutomatedSnapshotRetentionPeriod"),
        "redshift_manual_snapshot_retention_period": S("ManualSnapshotRetentionPeriod"),
        "redshift_cluster_security_groups": S("ClusterSecurityGroups", default=[])
        >> ForallBend(AwsRedshiftClusterSecurityGroupMembership.mapping),
        "redshift_vpc_security_groups": S("VpcSecurityGroups", default=[])
        >> ForallBend(AwsRedshiftVpcSecurityGroupMembership.mapping),
        "redshift_cluster_parameter_groups": S("ClusterParameterGroups", default=[])
        >> ForallBend(AwsRedshiftClusterParameterGroupStatus.mapping),
        "redshift_cluster_subnet_group_name": S("ClusterSubnetGroupName"),
        "redshift_vpc_id": S("VpcId"),
        "redshift_availability_zone": S("AvailabilityZone"),
        "redshift_preferred_maintenance_window": S("PreferredMaintenanceWindow"),
        "redshift_pending_modified_values": S("PendingModifiedValues")
        >> Bend(AwsRedshiftPendingModifiedValues.mapping),
        "redshift_cluster_version": S("ClusterVersion"),
        "redshift_allow_version_upgrade": S("AllowVersionUpgrade"),
        "redshift_number_of_nodes": S("NumberOfNodes"),
        "redshift_publicly_accessible": S("PubliclyAccessible"),
        "redshift_encrypted": S("Encrypted"),
        "redshift_restore_status": S("RestoreStatus") >> Bend(AwsRedshiftRestoreStatus.mapping),
        "redshift_data_transfer_progress": S("DataTransferProgress") >> Bend(AwsRedshiftDataTransferProgress.mapping),
        "redshift_hsm_status": S("HsmStatus") >> Bend(AwsRedshiftHsmStatus.mapping),
        "redshift_cluster_snapshot_copy_status": S("ClusterSnapshotCopyStatus")
        >> Bend(AwsRedshiftClusterSnapshotCopyStatus.mapping),
        "redshift_cluster_public_key": S("ClusterPublicKey"),
        "redshift_cluster_nodes": S("ClusterNodes", default=[]) >> ForallBend(AwsRedshiftClusterNode.mapping),
        "redshift_elastic_ip_status": S("ElasticIpStatus") >> Bend(AwsRedshiftElasticIpStatus.mapping),
        "redshift_cluster_revision_number": S("ClusterRevisionNumber"),
        "redshift_kms_key_id": S("KmsKeyId"),
        "redshift_enhanced_vpc_routing": S("EnhancedVpcRouting"),
        "redshift_iam_roles": S("IamRoles", default=[]) >> ForallBend(AwsRedshiftClusterIamRole.mapping),
        "redshift_pending_actions": S("PendingActions", default=[]),
        "redshift_maintenance_track_name": S("MaintenanceTrackName"),
        "redshift_elastic_resize_number_of_node_options": S("ElasticResizeNumberOfNodeOptions"),
        "redshift_deferred_maintenance_windows": S("DeferredMaintenanceWindows", default=[])
        >> ForallBend(AwsRedshiftDeferredMaintenanceWindow.mapping),
        "redshift_snapshot_schedule_identifier": S("SnapshotScheduleIdentifier"),
        "redshift_snapshot_schedule_state": S("SnapshotScheduleState"),
        "redshift_expected_next_snapshot_schedule_time": S("ExpectedNextSnapshotScheduleTime"),
        "redshift_expected_next_snapshot_schedule_time_status": S("ExpectedNextSnapshotScheduleTimeStatus"),
        "redshift_next_maintenance_window_start_time": S("NextMaintenanceWindowStartTime"),
        "redshift_resize_info": S("ResizeInfo") >> Bend(AwsRedshiftResizeInfo.mapping),
        "redshift_availability_zone_relocation_status": S("AvailabilityZoneRelocationStatus"),
        "redshift_cluster_namespace_arn": S("ClusterNamespaceArn"),
        "redshift_total_storage_capacity_in_mega_bytes": S("TotalStorageCapacityInMegaBytes"),
        "redshift_aqua_configuration": S("AquaConfiguration") >> Bend(AwsRedshiftAquaConfiguration.mapping),
        "redshift_default_iam_role_arn": S("DefaultIamRoleArn"),
        "redshift_reserved_node_exchange_status": S("ReservedNodeExchangeStatus")
        >> Bend(AwsRedshiftReservedNodeExchangeStatus.mapping),
    }
    redshift_node_type: Optional[str] = field(default=None)
    redshift_cluster_status: Optional[str] = field(default=None)
    redshift_cluster_availability_status: Optional[str] = field(default=None)
    redshift_modify_status: Optional[str] = field(default=None)
    redshift_master_username: Optional[str] = field(default=None)
    redshift_db_name: Optional[str] = field(default=None)
    redshift_endpoint: Optional[AwsRedshiftEndpoint] = field(default=None)
    redshift_automated_snapshot_retention_period: Optional[int] = field(default=None)
    redshift_manual_snapshot_retention_period: Optional[int] = field(default=None)
    redshift_cluster_security_groups: List[AwsRedshiftClusterSecurityGroupMembership] = field(factory=list)
    redshift_vpc_security_groups: List[AwsRedshiftVpcSecurityGroupMembership] = field(factory=list)
    redshift_cluster_parameter_groups: List[AwsRedshiftClusterParameterGroupStatus] = field(factory=list)
    redshift_cluster_subnet_group_name: Optional[str] = field(default=None)
    redshift_vpc_id: Optional[str] = field(default=None)
    redshift_availability_zone: Optional[str] = field(default=None)
    redshift_preferred_maintenance_window: Optional[str] = field(default=None)
    redshift_pending_modified_values: Optional[AwsRedshiftPendingModifiedValues] = field(default=None)
    redshift_cluster_version: Optional[str] = field(default=None)
    redshift_allow_version_upgrade: Optional[bool] = field(default=None)
    redshift_number_of_nodes: Optional[int] = field(default=None)
    redshift_publicly_accessible: Optional[bool] = field(default=None)
    redshift_encrypted: Optional[bool] = field(default=None)
    redshift_restore_status: Optional[AwsRedshiftRestoreStatus] = field(default=None)
    redshift_data_transfer_progress: Optional[AwsRedshiftDataTransferProgress] = field(default=None)
    redshift_hsm_status: Optional[AwsRedshiftHsmStatus] = field(default=None)
    redshift_cluster_snapshot_copy_status: Optional[AwsRedshiftClusterSnapshotCopyStatus] = field(default=None)
    redshift_cluster_public_key: Optional[str] = field(default=None)
    redshift_cluster_nodes: List[AwsRedshiftClusterNode] = field(factory=list)
    redshift_elastic_ip_status: Optional[AwsRedshiftElasticIpStatus] = field(default=None)
    redshift_cluster_revision_number: Optional[str] = field(default=None)
    redshift_kms_key_id: Optional[str] = field(default=None)
    redshift_enhanced_vpc_routing: Optional[bool] = field(default=None)
    redshift_iam_roles: List[AwsRedshiftClusterIamRole] = field(factory=list)
    redshift_pending_actions: List[str] = field(factory=list)
    redshift_maintenance_track_name: Optional[str] = field(default=None)
    redshift_elastic_resize_number_of_node_options: Optional[str] = field(default=None)
    redshift_deferred_maintenance_windows: List[AwsRedshiftDeferredMaintenanceWindow] = field(factory=list)
    redshift_snapshot_schedule_identifier: Optional[str] = field(default=None)
    redshift_snapshot_schedule_state: Optional[str] = field(default=None)
    redshift_expected_next_snapshot_schedule_time: Optional[datetime] = field(default=None)
    redshift_expected_next_snapshot_schedule_time_status: Optional[str] = field(default=None)
    redshift_next_maintenance_window_start_time: Optional[datetime] = field(default=None)
    redshift_resize_info: Optional[AwsRedshiftResizeInfo] = field(default=None)
    redshift_availability_zone_relocation_status: Optional[str] = field(default=None)
    redshift_cluster_namespace_arn: Optional[str] = field(default=None)
    redshift_total_storage_capacity_in_mega_bytes: Optional[int] = field(default=None)
    redshift_aqua_configuration: Optional[AwsRedshiftAquaConfiguration] = field(default=None)
    redshift_default_iam_role_arn: Optional[str] = field(default=None)
    redshift_reserved_node_exchange_status: Optional[AwsRedshiftReservedNodeExchangeStatus] = field(default=None)

    def connect_in_graph(self, builder: GraphBuilder, source: Json) -> None:
        if self.redshift_vpc_id:
            builder.dependant_node(self, reverse=True, clazz=AwsEc2Vpc, id=self.redshift_vpc_id)

        for vsg in self.redshift_vpc_security_groups:
            if vsg.vpc_security_group_id:
                builder.dependant_node(self, reverse=True, clazz=AwsEc2SecurityGroup, id=vsg.vpc_security_group_id)

        for role in self.redshift_iam_roles:
            if role.iam_role_arn:
                builder.dependant_node(self, reverse=True, clazz=AwsIamRole, arn=role.iam_role_arn)

        if self.redshift_cluster_subnet_group_name:
            builder.dependant_node(self, reverse=True, clazz=AwsEc2Subnet, name=self.redshift_cluster_subnet_group_name)

    def update_resource_tag(self, client: AwsClient, key: str, value: str) -> bool:
        client.call(
            service=self.api_spec.service,
            action="create_tags",
            result_name=None,
            ResourceName=self.arn,
            Tags=[{"Key": key, "Value": value}],
        )
        return True

    def delete_resource_tag(self, client: AwsClient, key: str) -> bool:
        client.call(
            service=self.api_spec.service,
            action="delete_tags",
            result_name=None,
            ResourceName=self.arn,
            TagKeys=[key],
        )
        return True

    def delete_resource(self, client: AwsClient) -> bool:
        client.call(
            service=self.api_spec.service,
            action="delete_cluster",
            result_name=None,
            ClusterIdentifier=self.id,
            SkipFinalClusterSnapshot=True,
        )
        return True


resources: List[Type[AwsResource]] = [AwsRedshiftCluster]



import json
from datetime import datetime
from functools import lru_cache
from typing import ClassVar, Dict, Optional, List, Any

from attr import field, frozen
from botocore.loaders import Loader

from resoto_plugin_aws.aws_client import AwsClient
from resotolib.json import from_json
from resotolib.json_bender import Bender, S, Bend, MapDict, F, ForallBend, bend
from resotolib.types import Json

EBS_TO_PRICING_NAMES = {
    "standard": "Magnetic",
    "gp2": "General Purpose",
    "gp3": "General Purpose",
    "io1": "Provisioned IOPS",
    "st1": "Throughput Optimized HDD",
    "sc1": "Cold HDD",
}


@lru_cache(maxsize=None)
def pricing_region(region: str) -> str:
    endpoints = Loader().load_data("endpoints")
    name: Optional[str] = bend(S("partitions")[0] >> S("regions", region, "description"), endpoints)
    if name is None:
        raise ValueError(f"Unknown pricing region: {region}")
    return name.replace("Europe", "EU")  # note: Europe is named differently in the price list


@frozen(eq=False)
class AwsPricingProduct:
    kind: ClassVar[str] = "aws_pricing_product"
    mapping: ClassVar[Dict[str, Bender]] = {
        "product_family": S("productFamily"),
        "sku": S("sku"),
        "attributes": S("attributes"),
    }
    product_family: Optional[str] = None
    attributes: Dict[str, str] = field(factory=dict)
    sku: Optional[str] = None


@frozen(eq=False)
class AwsPricingPriceDimension:
    kind: ClassVar[str] = "aws_pricing_price_dimension"
    mapping: ClassVar[Dict[str, Bender]] = {
        "unit": S("unit"),
        "end_range": S("endRange"),
        "description": S("description"),
        "applies_to": S("appliesTo"),
        "rate_code": S("rateCode"),
        "begin_range": S("beginRange"),
        "price_per_unit": S("pricePerUnit") >> MapDict(value_bender=F(float)),
    }
    unit: Optional[str] = None
    end_range: Optional[str] = None
    description: Optional[str] = None
    applies_to: List[Any] = field(factory=list)
    rate_code: Optional[str] = None
    begin_range: Optional[str] = None
    price_per_unit: Dict[str, float] = field(factory=dict)


@frozen(eq=False)
class AwsPricingTerm:
    kind: ClassVar[str] = "aws_pricing_term"
    mapping: ClassVar[Dict[str, Bender]] = {
        "sku": S("sku"),
        "effective_date": S("effectiveDate"),
        "offer_term_code": S("offerTermCode"),
        "term_attributes": S("termAttributes"),
        "price_dimensions": S("priceDimensions")
        >> F(lambda x: list(x.values()))
        >> ForallBend(AwsPricingPriceDimension.mapping),
    }
    sku: Optional[str] = None
    effective_date: Optional[datetime] = None
    offer_term_code: Optional[str] = None
    term_attributes: Dict[str, str] = field(factory=dict)
    price_dimensions: List[AwsPricingPriceDimension] = field(factory=list)


@frozen(eq=False)
class AwsPricingPrice:
    kind: ClassVar[str] = "aws_pricing_price"
    mapping: ClassVar[Dict[str, Bender]] = {
        "product": S("product") >> Bend(AwsPricingProduct.mapping),
        "service_code": S("serviceCode"),
        "terms": S("terms")
        >> MapDict(
            value_bender=F(lambda x: list(x.values()) if isinstance(x, dict) else [])
            >> ForallBend(AwsPricingTerm.mapping)
        ),
    }
    product: Optional[AwsPricingProduct] = None
    service_code: Optional[str] = None
    terms: Dict[str, List[AwsPricingTerm]] = field(factory=dict)

    @property
    def on_demand_price_usd(self) -> float:
        if terms := self.terms.get("OnDemand", []):
            if dim := terms[0].price_dimensions:
                return dim[0].price_per_unit.get("USD", 0)
        return 0

    @classmethod
    def single_price_for(
        cls, client: AwsClient, service_code: str, search_filter: List[Json]
    ) -> "Optional[AwsPricingPrice]":
        # Prices are only available in us-east-1
        prices = client.global_region.list(
            "pricing", "get-products", "PriceList", ServiceCode=service_code, Filters=search_filter, MaxResults=1
        )
        return from_json(bend(cls.mapping, json.loads(prices[0])), AwsPricingPrice) if prices else None

    @classmethod
    def volume_type_price(cls, client: AwsClient, volume_type: str, region: str) -> "Optional[AwsPricingPrice]":
        if volume_type not in EBS_TO_PRICING_NAMES:
            return None
        search_filter = [
            {"Type": "TERM_MATCH", "Field": "volumeType", "Value": EBS_TO_PRICING_NAMES[volume_type]},
            {"Type": "TERM_MATCH", "Field": "volumeApiName", "Value": volume_type},
            {"Type": "TERM_MATCH", "Field": "location", "Value": pricing_region(region)},
        ]
        return cls.single_price_for(client, "AmazonEC2", search_filter)

    @classmethod
    def instance_type_price(cls, client: AwsClient, instance_type: str, region: str) -> "Optional[AwsPricingPrice]":
        search_filter = [
            {"Type": "TERM_MATCH", "Field": "operatingSystem", "Value": "Linux"},
            {"Type": "TERM_MATCH", "Field": "operation", "Value": "RunInstances"},
            {"Type": "TERM_MATCH", "Field": "capacitystatus", "Value": "Used"},
            {"Type": "TERM_MATCH", "Field": "tenancy", "Value": "Shared"},
            {"Type": "TERM_MATCH", "Field": "instanceType", "Value": instance_type},
            {"Type": "TERM_MATCH", "Field": "location", "Value": pricing_region(region)},
        ]
        return cls.single_price_for(client, "AmazonEC2", search_filter)



from resoto_plugin_aws.resource.redshift import AwsRedshiftCluster
from test.resources import round_trip_for
from typing import Any, cast
from types import SimpleNamespace
from resoto_plugin_aws.aws_client import AwsClient


def test_redshift_cluster() -> None:
    res, builder = round_trip_for(AwsRedshiftCluster)
    assert len(builder.resources_of(AwsRedshiftCluster)) == 1
    assert len(res.tags) == 1


def test_tagging() -> None:
    cluster, _ = round_trip_for(AwsRedshiftCluster)

    def validate_update_args(**kwargs: Any) -> None:
        assert kwargs["action"] == "create_tags"
        assert kwargs["ResourceName"] == cluster.arn
        assert kwargs["Tags"] == [{"Key": "foo", "Value": "bar"}]

    def validate_delete_args(**kwargs: Any) -> None:
        assert kwargs["action"] == "delete_tags"
        assert kwargs["ResourceName"] == cluster.arn
        assert kwargs["TagKeys"] == ["foo"]

    client = cast(AwsClient, SimpleNamespace(call=validate_update_args))
    cluster.update_resource_tag(client, "foo", "bar")

    client = cast(AwsClient, SimpleNamespace(call=validate_delete_args))
    cluster.delete_resource_tag(client, "foo")


def test_deletion() -> None:
    cluster, _ = round_trip_for(AwsRedshiftCluster)

    def validate_delete_args(**kwargs: Any) -> None:
        assert kwargs["action"] == "delete_cluster"
        assert kwargs["ClusterIdentifier"] == cluster.id
        assert kwargs["SkipFinalClusterSnapshot"] is True

    client = cast(AwsClient, SimpleNamespace(call=validate_delete_args))
    cluster.delete_resource(client)



from typing import ClassVar, Dict, List, Type, Optional  # noqa: F401

from attrs import define
import botocore.exceptions
from resoto_plugin_aws.resource.base import AwsResource, AwsApiSpec
from resotolib.baseresources import BaseBucket, BaseAccount  # noqa: F401
from resotolib.json_bender import Bender, S
from resoto_plugin_aws.aws_client import AwsClient
from resoto_plugin_aws.utils import tags_as_dict


@define(eq=False, slots=False)
class AwsS3Bucket(AwsResource, BaseBucket):
    kind: ClassVar[str] = "aws_s3_bucket"
    api_spec: ClassVar[AwsApiSpec] = AwsApiSpec("s3", "list-buckets", "Buckets")
    mapping: ClassVar[Dict[str, Bender]] = {"id": S("Name"), "name": S("Name"), "ctime": S("CreationDate")}

    def _set_tags(self, client: AwsClient, tags: Dict[str, str]) -> bool:
        tag_set = [{"Key": k, "Value": v} for k, v in tags.items()]
        client.call(
            service="s3", action="put-bucket-tagging", result_name=None, Bucket=self.name, Tagging={"TagSet": tag_set}
        )
        return True

    def _get_tags(self, client: AwsClient) -> Dict[str, str]:
        """Fetch the S3 buckets tags from the AWS API."""
        tags: Dict[str, str] = {}
        try:
            response = client.call(service="s3", action="get_bucket_tagging", result_name="TagSet", Bucket=self.name)
            tags = tags_as_dict(response)  # type: ignore
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] != "NoSuchTagSet":
                raise
        return tags

    def update_resource_tag(self, client: AwsClient, key: str, value: str) -> bool:
        tags = self._get_tags(client)
        tags[key] = value
        return self._set_tags(client, tags)

    def delete_resource_tag(self, client: AwsClient, key: str) -> bool:
        tags = self._get_tags(client)
        if key in tags:
            del tags[key]
        else:
            raise KeyError(key)
        return self._set_tags(client, tags)

    def delete_resource(self, client: AwsClient) -> None:
        client.call(service="s3", action="delete_bucket", result_name=None, Bucket=self.name)


resources: List[Type[AwsResource]] = [AwsS3Bucket]



from datetime import datetime
from typing import ClassVar, Dict, List, Optional, Type
from attr import define, field
from resoto_plugin_aws.resource.base import AwsApiSpec, AwsResource, GraphBuilder
from resoto_plugin_aws.resource.cloudwatch import AwsCloudwatchQuery, AwsCloudwatchMetricData
from resoto_plugin_aws.resource.ec2 import AwsEc2SecurityGroup, AwsEc2Subnet, AwsEc2Vpc
from resoto_plugin_aws.utils import ToDict
from resotolib.baseresources import BaseAccount, BaseDatabase, ModelReference  # noqa: F401
from resotolib.json_bender import F, K, S, Bend, Bender, ForallBend
from resotolib.types import Json
from resotolib.utils import utc
from resoto_plugin_aws.aws_client import AwsClient


# todo: annotate with no serialization annotation
class RdsTaggable:
    def update_resource_tag(self, client: AwsClient, key: str, value: str) -> bool:
        if isinstance(self, AwsResource):
            if spec := self.api_spec:
                client.call(
                    service=spec.service,
                    action="add_tags_to_resource",
                    result_name=None,
                    ResourceName=self.arn,
                    Tags=[{"Key": key, "Value": value}],
                )
                return True
            return False
        return False

    def delete_resource_tag(self, client: AwsClient, key: str) -> bool:
        if isinstance(self, AwsResource):
            if spec := self.api_spec:
                client.call(
                    service=spec.service,
                    action="remove_tags_from_resource",
                    result_name=None,
                    ResourceName=self.arn,
                    TagKeys=[key],
                )
                return True
            return False
        return False


@define(eq=False, slots=False)
class AwsRdsEndpoint:
    kind: ClassVar[str] = "aws_rds_endpoint"
    mapping: ClassVar[Dict[str, Bender]] = {
        "address": S("Address"),
        "port": S("Port"),
        "hosted_zone_id": S("HostedZoneId"),
    }
    address: Optional[str] = field(default=None)
    port: Optional[int] = field(default=None)
    hosted_zone_id: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsRdsDBSecurityGroupMembership:
    kind: ClassVar[str] = "aws_rds_db_security_group_membership"
    mapping: ClassVar[Dict[str, Bender]] = {"db_security_group_name": S("DBSecurityGroupName"), "status": S("Status")}
    db_security_group_name: Optional[str] = field(default=None)
    status: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsRdsVpcSecurityGroupMembership:
    kind: ClassVar[str] = "aws_rds_vpc_security_group_membership"
    mapping: ClassVar[Dict[str, Bender]] = {"vpc_security_group_id": S("VpcSecurityGroupId"), "status": S("Status")}
    vpc_security_group_id: Optional[str] = field(default=None)
    status: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsRdsDBParameterGroupStatus:
    kind: ClassVar[str] = "aws_rds_db_parameter_group_status"
    mapping: ClassVar[Dict[str, Bender]] = {
        "db_parameter_group_name": S("DBParameterGroupName"),
        "parameter_apply_status": S("ParameterApplyStatus"),
    }
    db_parameter_group_name: Optional[str] = field(default=None)
    parameter_apply_status: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsRdsSubnet:
    kind: ClassVar[str] = "aws_rds_subnet"
    mapping: ClassVar[Dict[str, Bender]] = {
        "subnet_identifier": S("SubnetIdentifier"),
        "subnet_availability_zone": S("SubnetAvailabilityZone", "Name"),
        "subnet_outpost": S("SubnetOutpost", "Arn"),
        "subnet_status": S("SubnetStatus"),
    }
    subnet_identifier: Optional[str] = field(default=None)
    subnet_availability_zone: Optional[str] = field(default=None)
    subnet_outpost: Optional[str] = field(default=None)
    subnet_status: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsRdsDBSubnetGroup:
    kind: ClassVar[str] = "aws_rds_db_subnet_group"
    mapping: ClassVar[Dict[str, Bender]] = {
        "db_subnet_group_name": S("DBSubnetGroupName"),
        "db_subnet_group_description": S("DBSubnetGroupDescription"),
        "vpc_id": S("VpcId"),
        "subnet_group_status": S("SubnetGroupStatus"),
        "subnets": S("Subnets", default=[]) >> ForallBend(AwsRdsSubnet.mapping),
        "db_subnet_group_arn": S("DBSubnetGroupArn"),
        "supported_network_types": S("SupportedNetworkTypes", default=[]),
    }
    db_subnet_group_name: Optional[str] = field(default=None)
    db_subnet_group_description: Optional[str] = field(default=None)
    vpc_id: Optional[str] = field(default=None)
    subnet_group_status: Optional[str] = field(default=None)
    subnets: List[AwsRdsSubnet] = field(factory=list)
    db_subnet_group_arn: Optional[str] = field(default=None)
    supported_network_types: List[str] = field(factory=list)


@define(eq=False, slots=False)
class AwsRdsPendingCloudwatchLogsExports:
    kind: ClassVar[str] = "aws_rds_pending_cloudwatch_logs_exports"
    mapping: ClassVar[Dict[str, Bender]] = {
        "log_types_to_enable": S("LogTypesToEnable", default=[]),
        "log_types_to_disable": S("LogTypesToDisable", default=[]),
    }
    log_types_to_enable: List[str] = field(factory=list)
    log_types_to_disable: List[str] = field(factory=list)


@define(eq=False, slots=False)
class AwsRdsProcessorFeature:
    kind: ClassVar[str] = "aws_rds_processor_feature"
    mapping: ClassVar[Dict[str, Bender]] = {"name": S("Name"), "value": S("Value")}
    name: Optional[str] = field(default=None)
    value: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsRdsPendingModifiedValues:
    kind: ClassVar[str] = "aws_rds_pending_modified_values"
    mapping: ClassVar[Dict[str, Bender]] = {
        "db_instance_class": S("DBInstanceClass"),
        "allocated_storage": S("AllocatedStorage"),
        "master_user_password": S("MasterUserPassword"),
        "port": S("Port"),
        "backup_retention_period": S("BackupRetentionPeriod"),
        "multi_az": S("MultiAZ"),
        "engine_version": S("EngineVersion"),
        "license_model": S("LicenseModel"),
        "iops": S("Iops"),
        "db_instance_identifier": S("DBInstanceIdentifier"),
        "storage_type": S("StorageType"),
        "ca_certificate_identifier": S("CACertificateIdentifier"),
        "db_subnet_group_name": S("DBSubnetGroupName"),
        "pending_cloudwatch_logs_exports": S("PendingCloudwatchLogsExports")
        >> Bend(AwsRdsPendingCloudwatchLogsExports.mapping),
        "processor_features": S("ProcessorFeatures", default=[]) >> ForallBend(AwsRdsProcessorFeature.mapping),
        "iam_database_authentication_enabled": S("IAMDatabaseAuthenticationEnabled"),
        "automation_mode": S("AutomationMode"),
        "resume_full_automation_mode_time": S("ResumeFullAutomationModeTime"),
    }
    db_instance_class: Optional[str] = field(default=None)
    allocated_storage: Optional[int] = field(default=None)
    master_user_password: Optional[str] = field(default=None)
    port: Optional[int] = field(default=None)
    backup_retention_period: Optional[int] = field(default=None)
    multi_az: Optional[bool] = field(default=None)
    engine_version: Optional[str] = field(default=None)
    license_model: Optional[str] = field(default=None)
    iops: Optional[int] = field(default=None)
    db_instance_identifier: Optional[str] = field(default=None)
    storage_type: Optional[str] = field(default=None)
    ca_certificate_identifier: Optional[str] = field(default=None)
    db_subnet_group_name: Optional[str] = field(default=None)
    pending_cloudwatch_logs_exports: Optional[AwsRdsPendingCloudwatchLogsExports] = field(default=None)
    processor_features: List[AwsRdsProcessorFeature] = field(factory=list)
    iam_database_authentication_enabled: Optional[bool] = field(default=None)
    automation_mode: Optional[str] = field(default=None)
    resume_full_automation_mode_time: Optional[datetime] = field(default=None)


@define(eq=False, slots=False)
class AwsRdsOptionGroupMembership:
    kind: ClassVar[str] = "aws_rds_option_group_membership"
    mapping: ClassVar[Dict[str, Bender]] = {"option_group_name": S("OptionGroupName"), "status": S("Status")}
    option_group_name: Optional[str] = field(default=None)
    status: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsRdsDBInstanceStatusInfo:
    kind: ClassVar[str] = "aws_rds_db_instance_status_info"
    mapping: ClassVar[Dict[str, Bender]] = {
        "status_type": S("StatusType"),
        "normal": S("Normal"),
        "status": S("Status"),
        "message": S("Message"),
    }
    status_type: Optional[str] = field(default=None)
    normal: Optional[bool] = field(default=None)
    status: Optional[str] = field(default=None)
    message: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsRdsDomainMembership:
    kind: ClassVar[str] = "aws_rds_domain_membership"
    mapping: ClassVar[Dict[str, Bender]] = {
        "domain": S("Domain"),
        "status": S("Status"),
        "fqdn": S("FQDN"),
        "iam_role_name": S("IAMRoleName"),
    }
    domain: Optional[str] = field(default=None)
    status: Optional[str] = field(default=None)
    fqdn: Optional[str] = field(default=None)
    iam_role_name: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsRdsDBInstanceRole:
    kind: ClassVar[str] = "aws_rds_db_instance_role"
    mapping: ClassVar[Dict[str, Bender]] = {
        "role_arn": S("RoleArn"),
        "feature_name": S("FeatureName"),
        "status": S("Status"),
    }
    role_arn: Optional[str] = field(default=None)
    feature_name: Optional[str] = field(default=None)
    status: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsRdsTag:
    kind: ClassVar[str] = "aws_rds_tag"
    mapping: ClassVar[Dict[str, Bender]] = {"key": S("Key"), "value": S("Value")}
    key: Optional[str] = field(default=None)
    value: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsRdsInstance(RdsTaggable, AwsResource, BaseDatabase):
    kind: ClassVar[str] = "aws_rds_instance"
    api_spec: ClassVar[AwsApiSpec] = AwsApiSpec("rds", "describe-db-instances", "DBInstances")
    reference_kinds: ClassVar[ModelReference] = {
        "predecessors": {
            "default": ["aws_vpc", "aws_ec2_security_group", "aws_ec2_subnet"],
            "delete": ["aws_vpc", "aws_ec2_security_group", "aws_ec2_subnet"],
        },
    }
    mapping: ClassVar[Dict[str, Bender]] = {
        "id": S("DBInstanceIdentifier"),
        "tags": S("TagList", default=[]) >> ForallBend(AwsRdsTag.mapping) >> ToDict(),
        "name": S("DBName"),
        "ctime": S("InstanceCreateTime"),
        "arn": S("DBInstanceArn"),
        "db_type": S("Engine"),
        "db_status": S("DBInstanceStatus"),
        "db_endpoint": S("Endpoint", "Address") + K(":") + (S("Endpoint", "Port") >> F(str)),
        "db_version": S("EngineVersion"),
        "db_publicly_accessible": S("PubliclyAccessible"),
        "instance_type": S("DBInstanceClass"),
        "volume_size": S("AllocatedStorage"),
        "volume_iops": S("Iops"),
        "volume_encrypted": S("StorageEncrypted"),
        "rds_automatic_restart_time": S("AutomaticRestartTime"),
        "rds_master_username": S("MasterUsername"),
        "rds_preferred_backup_window": S("PreferredBackupWindow"),
        "rds_backup_retention_period": S("BackupRetentionPeriod"),
        "rds_db_security_groups": S("DBSecurityGroups", default=[])
        >> ForallBend(AwsRdsDBSecurityGroupMembership.mapping),
        "rds_vpc_security_groups": S("VpcSecurityGroups", default=[])
        >> ForallBend(AwsRdsVpcSecurityGroupMembership.mapping),
        "rds_db_parameter_groups": S("DBParameterGroups", default=[])
        >> ForallBend(AwsRdsDBParameterGroupStatus.mapping),
        "rds_availability_zone": S("AvailabilityZone"),
        "rds_db_subnet_group": S("DBSubnetGroup") >> Bend(AwsRdsDBSubnetGroup.mapping),
        "rds_preferred_maintenance_window": S("PreferredMaintenanceWindow"),
        "rds_pending_modified_values": S("PendingModifiedValues") >> Bend(AwsRdsPendingModifiedValues.mapping),
        "rds_latest_restorable_time": S("LatestRestorableTime"),
        "rds_multi_az": S("MultiAZ"),
        "rds_auto_minor_version_upgrade": S("AutoMinorVersionUpgrade"),
        "rds_read_replica_source_db_instance_identifier": S("ReadReplicaSourceDBInstanceIdentifier"),
        "rds_read_replica_db_instance_identifiers": S("ReadReplicaDBInstanceIdentifiers", default=[]),
        "rds_read_replica_db_cluster_identifiers": S("ReadReplicaDBClusterIdentifiers", default=[]),
        "rds_replica_mode": S("ReplicaMode"),
        "rds_license_model": S("LicenseModel"),
        "rds_option_group_memberships": S("OptionGroupMemberships", default=[])
        >> ForallBend(AwsRdsOptionGroupMembership.mapping),
        "rds_character_set_name": S("CharacterSetName"),
        "rds_nchar_character_set_name": S("NcharCharacterSetName"),
        "rds_secondary_availability_zone": S("SecondaryAvailabilityZone"),
        "rds_status_infos": S("StatusInfos", default=[]) >> ForallBend(AwsRdsDBInstanceStatusInfo.mapping),
        "rds_storage_type": S("StorageType"),
        "rds_tde_credential_arn": S("TdeCredentialArn"),
        "rds_db_instance_port": S("DbInstancePort"),
        "rds_db_cluster_identifier": S("DBClusterIdentifier"),
        "rds_kms_key_id": S("KmsKeyId"),
        "rds_dbi_resource_id": S("DbiResourceId"),
        "rds_ca_certificate_identifier": S("CACertificateIdentifier"),
        "rds_domain_memberships": S("DomainMemberships", default=[]) >> ForallBend(AwsRdsDomainMembership.mapping),
        "rds_copy_tags_to_snapshot": S("CopyTagsToSnapshot"),
        "rds_monitoring_interval": S("MonitoringInterval"),
        "rds_enhanced_monitoring_resource_arn": S("EnhancedMonitoringResourceArn"),
        "rds_monitoring_role_arn": S("MonitoringRoleArn"),
        "rds_promotion_tier": S("PromotionTier"),
        "rds_timezone": S("Timezone"),
        "rds_iam_database_authentication_enabled": S("IAMDatabaseAuthenticationEnabled"),
        "rds_performance_insights_enabled": S("PerformanceInsightsEnabled"),
        "rds_performance_insights_kms_key_id": S("PerformanceInsightsKMSKeyId"),
        "rds_performance_insights_retention_period": S("PerformanceInsightsRetentionPeriod"),
        "rds_enabled_cloudwatch_logs_exports": S("EnabledCloudwatchLogsExports", default=[]),
        "rds_processor_features": S("ProcessorFeatures", default=[]) >> ForallBend(AwsRdsProcessorFeature.mapping),
        "rds_deletion_protection": S("DeletionProtection"),
        "rds_associated_roles": S("AssociatedRoles", default=[]) >> ForallBend(AwsRdsDBInstanceRole.mapping),
        "rds_listener_endpoint": S("ListenerEndpoint") >> Bend(AwsRdsEndpoint.mapping),
        "rds_max_allocated_storage": S("MaxAllocatedStorage"),
        "rds_db_instance_automated_backups_replications": S("DBInstanceAutomatedBackupsReplications", default=[])
        >> ForallBend(S("DBInstanceAutomatedBackupsArn")),
        "rds_customer_owned_ip_enabled": S("CustomerOwnedIpEnabled"),
        "rds_aws_backup_recovery_point_arn": S("AwsBackupRecoveryPointArn"),
        "rds_activity_stream_status": S("ActivityStreamStatus"),
        "rds_activity_stream_kms_key_id": S("ActivityStreamKmsKeyId"),
        "rds_activity_stream_kinesis_stream_name": S("ActivityStreamKinesisStreamName"),
        "rds_activity_stream_mode": S("ActivityStreamMode"),
        "rds_activity_stream_engine_native_audit_fields_included": S("ActivityStreamEngineNativeAuditFieldsIncluded"),
        "rds_automation_mode": S("AutomationMode"),
        "rds_resume_full_automation_mode_time": S("ResumeFullAutomationModeTime"),
        "rds_custom_iam_instance_profile": S("CustomIamInstanceProfile"),
        "rds_backup_target": S("BackupTarget"),
        "rds_network_type": S("NetworkType"),
    }
    arn: Optional[str] = field(default=None)
    db_type: str = field(default=None)
    db_status: str = field(default=None)
    db_endpoint: str = field(default=None)
    db_version: Optional[str] = field(default=None)
    db_publicly_accessible: Optional[bool] = field(default=None)
    instance_type: str = field(default=None)
    volume_size: int = field(default=None)
    volume_iops: Optional[int] = field(default=None)
    volume_encrypted: Optional[bool] = field(default=None)
    rds_automatic_restart_time: Optional[datetime] = field(default=None)
    rds_master_username: Optional[str] = field(default=None)
    rds_preferred_backup_window: Optional[str] = field(default=None)
    rds_backup_retention_period: Optional[int] = field(default=None)
    rds_db_security_groups: List[AwsRdsDBSecurityGroupMembership] = field(factory=list)
    rds_vpc_security_groups: List[AwsRdsVpcSecurityGroupMembership] = field(factory=list)
    rds_db_parameter_groups: List[AwsRdsDBParameterGroupStatus] = field(factory=list)
    rds_availability_zone: Optional[str] = field(default=None)
    rds_db_subnet_group: Optional[AwsRdsDBSubnetGroup] = field(default=None)
    rds_preferred_maintenance_window: Optional[str] = field(default=None)
    rds_pending_modified_values: Optional[AwsRdsPendingModifiedValues] = field(default=None)
    rds_latest_restorable_time: Optional[datetime] = field(default=None)
    rds_multi_az: Optional[bool] = field(default=None)
    rds_auto_minor_version_upgrade: Optional[bool] = field(default=None)
    rds_read_replica_source_db_instance_identifier: Optional[str] = field(default=None)
    rds_read_replica_db_instance_identifiers: List[str] = field(factory=list)
    rds_read_replica_db_cluster_identifiers: List[str] = field(factory=list)
    rds_replica_mode: Optional[str] = field(default=None)
    rds_license_model: Optional[str] = field(default=None)
    rds_option_group_memberships: List[AwsRdsOptionGroupMembership] = field(factory=list)
    rds_character_set_name: Optional[str] = field(default=None)
    rds_nchar_character_set_name: Optional[str] = field(default=None)
    rds_secondary_availability_zone: Optional[str] = field(default=None)
    rds_status_infos: List[AwsRdsDBInstanceStatusInfo] = field(factory=list)
    rds_storage_type: Optional[str] = field(default=None)
    rds_tde_credential_arn: Optional[str] = field(default=None)
    rds_db_instance_port: Optional[int] = field(default=None)
    rds_db_cluster_identifier: Optional[str] = field(default=None)
    rds_kms_key_id: Optional[str] = field(default=None)
    rds_dbi_resource_id: Optional[str] = field(default=None)
    rds_ca_certificate_identifier: Optional[str] = field(default=None)
    rds_domain_memberships: List[AwsRdsDomainMembership] = field(factory=list)
    rds_copy_tags_to_snapshot: Optional[bool] = field(default=None)
    rds_monitoring_interval: Optional[int] = field(default=None)
    rds_enhanced_monitoring_resource_arn: Optional[str] = field(default=None)
    rds_monitoring_role_arn: Optional[str] = field(default=None)
    rds_promotion_tier: Optional[int] = field(default=None)
    rds_timezone: Optional[str] = field(default=None)
    rds_iam_database_authentication_enabled: Optional[bool] = field(default=None)
    rds_performance_insights_enabled: Optional[bool] = field(default=None)
    rds_performance_insights_kms_key_id: Optional[str] = field(default=None)
    rds_performance_insights_retention_period: Optional[int] = field(default=None)
    rds_enabled_cloudwatch_logs_exports: List[str] = field(factory=list)
    rds_processor_features: List[AwsRdsProcessorFeature] = field(factory=list)
    rds_deletion_protection: Optional[bool] = field(default=None)
    rds_associated_roles: List[AwsRdsDBInstanceRole] = field(factory=list)
    rds_listener_endpoint: Optional[AwsRdsEndpoint] = field(default=None)
    rds_max_allocated_storage: Optional[int] = field(default=None)
    rds_db_instance_automated_backups_replications: List[str] = field(factory=list)
    rds_customer_owned_ip_enabled: Optional[bool] = field(default=None)
    rds_aws_backup_recovery_point_arn: Optional[str] = field(default=None)
    rds_activity_stream_status: Optional[str] = field(default=None)
    rds_activity_stream_kms_key_id: Optional[str] = field(default=None)
    rds_activity_stream_kinesis_stream_name: Optional[str] = field(default=None)
    rds_activity_stream_mode: Optional[str] = field(default=None)
    rds_activity_stream_engine_native_audit_fields_included: Optional[bool] = field(default=None)
    rds_automation_mode: Optional[str] = field(default=None)
    rds_resume_full_automation_mode_time: Optional[datetime] = field(default=None)
    rds_custom_iam_instance_profile: Optional[str] = field(default=None)
    rds_backup_target: Optional[str] = field(default=None)
    rds_network_type: Optional[str] = field(default=None)

    @classmethod
    def collect(cls: Type[AwsResource], json: List[Json], builder: GraphBuilder) -> None:
        instances: List[AwsRdsInstance] = []

        def update_atime_mtime() -> None:
            delta = builder.config.atime_mtime_granularity()
            queries = []
            now = utc()
            start = now - builder.config.atime_mtime_period()
            lookup: Dict[str, AwsRdsInstance] = {}
            for rds in instances:
                vid = rds.id
                lookup[vid] = rds
                queries.append(
                    AwsCloudwatchQuery.create("DatabaseConnections", "AWS/RDS", delta, vid, DBInstanceIdentifier=vid)
                )

            for query, metric in AwsCloudwatchMetricData.query_for(builder.client, queries, start, now).items():
                if non_zero := metric.first_non_zero():
                    at, value = non_zero
                    rds = lookup[query.ref_id]
                    rds.atime = at
                    rds.mtime = at
                    lookup.pop(query.ref_id, None)

            # all volumes in this list do not have value in cloudwatch
            # fall back to either ctime or start time whatever is more recent.
            for rds in lookup.values():
                t = max(rds.ctime or start, start)
                rds.atime = t
                rds.mtime = t

        for js in json:
            instance = AwsRdsInstance.from_api(js)
            instances.append(instance)
            builder.add_node(instance, js)
        update_atime_mtime()

    def connect_in_graph(self, builder: GraphBuilder, source: Json) -> None:
        super().connect_in_graph(builder, source)
        for group in self.rds_vpc_security_groups:
            builder.dependant_node(
                self,
                reverse=True,
                delete_same_as_default=True,
                clazz=AwsEc2SecurityGroup,
                id=group.vpc_security_group_id,
            )
        if self.rds_db_subnet_group and self.rds_db_subnet_group.vpc_id:
            builder.dependant_node(
                self, reverse=True, delete_same_as_default=True, clazz=AwsEc2Vpc, id=self.rds_db_subnet_group.vpc_id
            )
        if self.rds_db_subnet_group:
            for subnet in self.rds_db_subnet_group.subnets:
                subnet_id = subnet.subnet_identifier
                builder.dependant_node(
                    self, reverse=True, delete_same_as_default=True, clazz=AwsEc2Subnet, id=subnet_id
                )

    def delete_resource(self, client: AwsClient) -> bool:
        client.call(
            service=self.api_spec.service,
            action="delete_db_instance",
            result_name=None,
            DBInstanceIdentifier=self.id,
            SkipFinalSnapshot=True,
        )
        return True


resources: List[Type[AwsResource]] = [AwsRdsInstance]



from resoto_plugin_aws.aws_client import AwsClient
from resoto_plugin_aws.resource.pricing import AwsPricingPrice, pricing_region
from resotolib.json import to_json, from_json
from test import aws_client, builder  # noqa: F401


def test_json_serialization(aws_client: AwsClient) -> None:
    price = AwsPricingPrice.instance_type_price(aws_client, "m4.large", "us-east-1")
    assert price
    again = from_json(to_json(price), AwsPricingPrice)
    assert to_json(price) == to_json(again)


def test_price_region() -> None:
    assert pricing_region("us-east-1") == "US East (N. Virginia)"
    assert pricing_region("eu-central-1") == "EU (Frankfurt)"


def test_instance_type_pricing(aws_client: AwsClient) -> None:
    price = AwsPricingPrice.instance_type_price(aws_client, "m4.large", "us-east-1")
    assert price is not None
    assert price.on_demand_price_usd == 0.051


def test_volume_type_pricing(aws_client: AwsClient) -> None:
    price = AwsPricingPrice.volume_type_price(aws_client, "gp2", "us-east-1")
    assert price is not None
    assert price.on_demand_price_usd == 0.119



from resoto_plugin_aws.resource.cloudformation import AwsCloudFormationStack, AwsCloudFormationStackSet
from resoto_plugin_aws.aws_client import AwsClient
from test.resources import round_trip_for
from typing import Any, cast
from types import SimpleNamespace
from functools import partial


def test_cloud_formation_stacks() -> None:
    round_trip_for(AwsCloudFormationStack)


def test_cloud_formation_stack_sets() -> None:
    round_trip_for(AwsCloudFormationStack)


def test_cloud_formation_stack_tagging() -> None:
    cf, _ = round_trip_for(AwsCloudFormationStack)

    tag = "alpha.eksctl.io/nodegroup-name"

    def validate_args(delete: bool, **kwargs: Any) -> Any:

        assert kwargs["action"] in {"describe_stacks", "update_stack"}
        if kwargs["action"] == "describe_stacks":
            assert kwargs["StackName"] == cf.name
            return [{"StackStatus": "complete"}]
        if kwargs["action"] == "update_stack":
            tags = cf.tags
            if delete:
                del tags[tag]
            else:
                tags.update({"foo": "bar"})
            assert kwargs["StackName"] == cf.name
            assert kwargs["Capabilities"] == ["CAPABILITY_NAMED_IAM"]
            assert kwargs["UsePreviousTemplate"] is True
            assert kwargs["Tags"] == [{"Key": label, "Value": value} for label, value in tags.items()]
            assert kwargs["Parameters"] == [
                {"ParameterKey": parameter, "UsePreviousValue": True} for parameter in cf.stack_parameters.keys()
            ]

    client = cast(AwsClient, SimpleNamespace(call=partial(validate_args, delete=False)))
    cf.update_resource_tag(client, "foo", "bar")

    client = cast(AwsClient, SimpleNamespace(call=partial(validate_args, delete=True)))
    cf.delete_resource_tag(client, tag)


def test_cloud_formation_stack_delete() -> None:
    cf, _ = round_trip_for(AwsCloudFormationStack)

    def validate_args(**kwargs: Any) -> Any:
        assert kwargs["action"] == "delete_stack"
        assert kwargs["StackName"] == cf.name

    client = cast(AwsClient, SimpleNamespace(call=partial(validate_args)))
    cf.delete_resource(client)


def test_cloud_formation_stack_set_tagging() -> None:
    cf, _ = round_trip_for(AwsCloudFormationStackSet)
    cf.tags["bar"] = "bar"

    def validate_args(delete: bool, **kwargs: Any) -> None:

        assert kwargs["action"] == "update_stack_set"
        tags = cf.tags
        if delete:
            del tags["bar"]
        else:
            tags.update({"foo": "bar"})
        assert kwargs["StackSetName"] == cf.name
        assert kwargs["Capabilities"] == ["CAPABILITY_NAMED_IAM"]
        assert kwargs["UsePreviousTemplate"] is True
        assert kwargs["Tags"] == [{"Key": label, "Value": value} for label, value in tags.items()]
        assert kwargs["Parameters"] == [
            {"ParameterKey": parameter, "UsePreviousValue": True}
            for parameter in (cf.stack_set_parameters or {}).keys()
        ]

    client = cast(AwsClient, SimpleNamespace(call=partial(validate_args, delete=False)))
    cf.update_resource_tag(client, "foo", "bar")

    client = cast(AwsClient, SimpleNamespace(call=partial(validate_args, delete=True)))
    cf.delete_resource_tag(client, "bar")


def test_cloud_formation_stack_set_delete() -> None:
    cf, _ = round_trip_for(AwsCloudFormationStackSet)

    def validate_args(**kwargs: Any) -> None:
        assert kwargs["action"] == "delete_stack_set"
        assert kwargs["StackSetName"] == cf.name

    client = cast(AwsClient, SimpleNamespace(call=partial(validate_args)))
    cf.delete_resource(client)



from resoto_plugin_aws.resources import AWSRegion
from resoto_plugin_aws.utils import arn_partition


def test_arn_partition() -> None:
    us_east_1 = AWSRegion(id="us-east-1")
    cn_north_1 = AWSRegion(id="cn-north-1")
    us_gov_east_1 = AWSRegion(id="us-gov-east-1")
    assert arn_partition(us_east_1) == "aws"
    assert arn_partition(cn_north_1) == "aws-cn"
    assert arn_partition(us_gov_east_1) == "aws-us-gov"



#!/usr/bin/env python
import sys
from prompt_toolkit.shortcuts import button_dialog
from resoto_plugin_aws.utils import aws_session
from resoto_plugin_aws.resources import AWSAccount
from resoto_plugin_aws import AWSPlugin, current_account_id
from resotolib.utils import make_valid_timestamp
from resotolib.args import get_arg_parser, ArgumentParser
import resotolib.logger
import re
from datetime import datetime

resotolib.logger.getLogger("resoto.cmd").setLevel(resotolib.logger.INFO)
log = resotolib.logger.getLogger("resoto.cmd")

argv = sys.argv[1:]
if "-v" in argv or "--verbose" in argv:
    resotolib.logger.getLogger("resoto.cmd").setLevel(resotolib.logger.DEBUG)
log.info("resoto S3 object purger")


def add_args(arg_parser: ArgumentParser) -> None:
    AWSPlugin.add_args(arg_parser)
    arg_parser.add_argument(
        "--aws-s3-bucket",
        help="AWS S3 Bucket (default: None)",
        dest="aws_s3_bucket",
        required=True,
        default=None,
    )
    arg_parser.add_argument(
        "--aws-s3-prefix",
        help="AWS S3 Prefix (default: None)",
        dest="aws_s3_prefix",
        type=str,
        default="",
    )
    arg_parser.add_argument(
        "--aws-s3-mtime",
        help="AWS S3 Mtime (default: None)",
        dest="aws_s3_mtime",
        type=str,
        default=None,
    )
    arg_parser.add_argument(
        "--aws-s3-pattern",
        help="AWS S3 Pattern (default: None)",
        dest="aws_s3_pattern",
        type=str,
        default=None,
    )
    arg_parser.add_argument(
        "--aws-s3-yes",
        help="AWS S3 delete without asking (default: False)",
        dest="aws_s3_yes",
        action="store_true",
        default=False,
    )


arg_parser = get_arg_parser()
add_args(arg_parser)
arg_parser.parse_args()


def main():
    if ArgumentParser.args.aws_role and ArgumentParser.args.aws_account:
        accounts = [
            AWSAccount(id=aws_account_id, tags={}, role=ArgumentParser.args.aws_role)
            for aws_account_id in ArgumentParser.args.aws_account
        ]
    else:
        accounts = [AWSAccount(id=current_account_id(), tags={})]

    if len(accounts) != 1:
        log.error("This tool only supports a single account at a time")
        sys.exit(1)

    account = accounts[0]
    session = aws_session(account.id, account.role)
    client = session.client("s3")
    bucket = ArgumentParser.args.aws_s3_bucket
    prefix = ArgumentParser.args.aws_s3_prefix

    mtime = (
        make_valid_timestamp(datetime.fromisoformat(ArgumentParser.args.aws_s3_mtime))
        if ArgumentParser.args.aws_s3_mtime
        else None
    )

    is_truncated = True
    max_keys = 500
    key_marker = None
    version_id_marker = None

    while is_truncated is True:
        if key_marker and version_id_marker:
            version_list = client.list_object_versions(
                Bucket=bucket,
                MaxKeys=max_keys,
                Prefix=prefix,
                KeyMarker=key_marker,
                VersionIdMarker=version_id_marker,
            )
        elif key_marker:
            version_list = client.list_object_versions(
                Bucket=bucket, MaxKeys=max_keys, Prefix=prefix, KeyMarker=key_marker
            )
        else:
            version_list = client.list_object_versions(Bucket=bucket, MaxKeys=max_keys, Prefix=prefix)
        is_truncated = version_list.get("IsTruncated", False)
        key_marker = version_list.get("NextKeyMarker")
        version_id_marker = version_list.get("NextVersionIdMarker")

        delete_objects = []
        versions = version_list.get("Versions", [])
        versions.extend(version_list.get("DeleteMarkers", []))
        for v in versions:
            object_version = v["VersionId"]
            object_key = v["Key"]
            # object_size = v["Size"]
            object_mtime = make_valid_timestamp(v["LastModified"])

            if mtime and object_mtime > mtime:
                log.debug(f"Object {object_key} with mtime {object_mtime} newer than mtime {mtime}")
                continue
            if (
                ArgumentParser.args.aws_s3_pattern
                and bool(re.search(ArgumentParser.args.aws_s3_pattern, str(object_key))) is False
            ):
                log.debug(f"Object {object_key} does not match {ArgumentParser.args.aws_s3_pattern}")
                continue
            log.info(
                (
                    f"Object {object_key} with version {object_version} and mtime"
                    f" {object_mtime} matches {ArgumentParser.args.aws_s3_pattern}"
                )
            )
            delete_objects.append({"VersionId": object_version, "Key": object_key})
        try:
            if len(delete_objects) > 0:
                str_delete_objects = "\n".join([do["Key"] for do in delete_objects])
                if ArgumentParser.args.aws_s3_yes is True:
                    confirm_delete = True
                else:
                    confirm_delete = button_dialog(
                        title=f"Delete {len(delete_objects)} S3 objects?",
                        text=f"Really delete these objects?\n{str_delete_objects}",
                        buttons=[("Yes", True), ("No", False), ("Abort", None)],
                    ).run()

                if confirm_delete is None:
                    sys.exit(0)
                elif confirm_delete is True:
                    response = client.delete_objects(Bucket=bucket, Delete={"Objects": delete_objects})
                    log.info(f"Delete response {response}")
        except Exception:
            log.exception("Something went wrong trying to delete")


if __name__ == "__main__":
    main()



from typing import ClassVar, Dict, Optional, Type, List

from attrs import define, field

from resoto_plugin_aws.resource.base import AwsResource, GraphBuilder, AwsApiSpec
from resoto_plugin_aws.resource.ec2 import AwsEc2Vpc, AwsEc2Subnet, AwsEc2Instance, AwsEc2SecurityGroup
from resoto_plugin_aws.utils import ToDict
from resotolib.baseresources import BaseLoadBalancer, EdgeType, BaseAccount, ModelReference  # noqa: F401
from resotolib.json import from_json
from resotolib.json_bender import Bender, S, Bend, bend, ForallBend, K
from resotolib.types import Json
from resoto_plugin_aws.aws_client import AwsClient


# todo: annotate with no serialization annotation
class ElbV2Taggable:
    def update_resource_tag(self, client: AwsClient, key: str, value: str) -> bool:
        if isinstance(self, AwsResource):
            if spec := self.api_spec:
                client.call(
                    service=spec.service,
                    action="add_tags",
                    result_name=None,
                    ResourceArns=[self.arn],
                    Tags=[{"Key": key, "Value": value}],
                )
                return True
            return False
        return False

    def delete_resource_tag(self, client: AwsClient, key: str) -> bool:
        if isinstance(self, AwsResource):
            if spec := self.api_spec:
                client.call(
                    service=spec.service,
                    action="remove_tags",
                    result_name=None,
                    ResourceArns=[self.arn],
                    TagKeys=[key],
                )
                return True
            return False
        return False


@define(eq=False, slots=False)
class AwsAlbLoadBalancerState:
    kind: ClassVar[str] = "aws_alb_load_balancer_state"
    mapping: ClassVar[Dict[str, Bender]] = {"code": S("Code"), "reason": S("Reason")}
    code: Optional[str] = field(default=None)
    reason: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsAlbLoadBalancerAddress:
    kind: ClassVar[str] = "aws_alb_load_balancer_address"
    mapping: ClassVar[Dict[str, Bender]] = {
        "ip_address": S("IpAddress"),
        "allocation_id": S("AllocationId"),
        "private_i_pv4_address": S("PrivateIPv4Address"),
        "i_pv6_address": S("IPv6Address"),
    }
    ip_address: Optional[str] = field(default=None)
    allocation_id: Optional[str] = field(default=None)
    private_i_pv4_address: Optional[str] = field(default=None)
    i_pv6_address: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsAlbAvailabilityZone:
    kind: ClassVar[str] = "aws_alb_availability_zone"
    mapping: ClassVar[Dict[str, Bender]] = {
        "zone_name": S("ZoneName"),
        "subnet_id": S("SubnetId"),
        "outpost_id": S("OutpostId"),
        "load_balancer_addresses": S("LoadBalancerAddresses", default=[])
        >> ForallBend(AwsAlbLoadBalancerAddress.mapping),
    }
    zone_name: Optional[str] = field(default=None)
    subnet_id: Optional[str] = field(default=None)
    outpost_id: Optional[str] = field(default=None)
    load_balancer_addresses: List[AwsAlbLoadBalancerAddress] = field(factory=list)


@define(eq=False, slots=False)
class AwsAlbCertificate:
    kind: ClassVar[str] = "aws_alb_certificate"
    mapping: ClassVar[Dict[str, Bender]] = {"certificate_arn": S("CertificateArn"), "is_default": S("IsDefault")}
    certificate_arn: Optional[str] = field(default=None)
    is_default: Optional[bool] = field(default=None)


@define(eq=False, slots=False)
class AwsAlbAuthenticateOidcActionConfig:
    kind: ClassVar[str] = "aws_alb_authenticate_oidc_action_config"
    mapping: ClassVar[Dict[str, Bender]] = {
        "issuer": S("Issuer"),
        "authorization_endpoint": S("AuthorizationEndpoint"),
        "token_endpoint": S("TokenEndpoint"),
        "user_info_endpoint": S("UserInfoEndpoint"),
        "client_id": S("ClientId"),
        "client_secret": S("ClientSecret"),
        "session_cookie_name": S("SessionCookieName"),
        "scope": S("Scope"),
        "session_timeout": S("SessionTimeout"),
        "authentication_request_extra_params": S("AuthenticationRequestExtraParams"),
        "on_unauthenticated_request": S("OnUnauthenticatedRequest"),
        "use_existing_client_secret": S("UseExistingClientSecret"),
    }
    issuer: Optional[str] = field(default=None)
    authorization_endpoint: Optional[str] = field(default=None)
    token_endpoint: Optional[str] = field(default=None)
    user_info_endpoint: Optional[str] = field(default=None)
    client_id: Optional[str] = field(default=None)
    client_secret: Optional[str] = field(default=None)
    session_cookie_name: Optional[str] = field(default=None)
    scope: Optional[str] = field(default=None)
    session_timeout: Optional[int] = field(default=None)
    authentication_request_extra_params: Optional[Dict[str, str]] = field(default=None)
    on_unauthenticated_request: Optional[str] = field(default=None)
    use_existing_client_secret: Optional[bool] = field(default=None)


@define(eq=False, slots=False)
class AwsAlbAuthenticateCognitoActionConfig:
    kind: ClassVar[str] = "aws_alb_authenticate_cognito_action_config"
    mapping: ClassVar[Dict[str, Bender]] = {
        "user_pool_arn": S("UserPoolArn"),
        "user_pool_client_id": S("UserPoolClientId"),
        "user_pool_domain": S("UserPoolDomain"),
        "session_cookie_name": S("SessionCookieName"),
        "scope": S("Scope"),
        "session_timeout": S("SessionTimeout"),
        "authentication_request_extra_params": S("AuthenticationRequestExtraParams"),
        "on_unauthenticated_request": S("OnUnauthenticatedRequest"),
    }
    user_pool_arn: Optional[str] = field(default=None)
    user_pool_client_id: Optional[str] = field(default=None)
    user_pool_domain: Optional[str] = field(default=None)
    session_cookie_name: Optional[str] = field(default=None)
    scope: Optional[str] = field(default=None)
    session_timeout: Optional[int] = field(default=None)
    authentication_request_extra_params: Optional[Dict[str, str]] = field(default=None)
    on_unauthenticated_request: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsAlbRedirectActionConfig:
    kind: ClassVar[str] = "aws_alb_redirect_action_config"
    mapping: ClassVar[Dict[str, Bender]] = {
        "protocol": S("Protocol"),
        "port": S("Port"),
        "host": S("Host"),
        "path": S("Path"),
        "query": S("Query"),
        "status_code": S("StatusCode"),
    }
    protocol: Optional[str] = field(default=None)
    port: Optional[str] = field(default=None)
    host: Optional[str] = field(default=None)
    path: Optional[str] = field(default=None)
    query: Optional[str] = field(default=None)
    status_code: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsAlbFixedResponseActionConfig:
    kind: ClassVar[str] = "aws_alb_fixed_response_action_config"
    mapping: ClassVar[Dict[str, Bender]] = {
        "message_body": S("MessageBody"),
        "status_code": S("StatusCode"),
        "content_type": S("ContentType"),
    }
    message_body: Optional[str] = field(default=None)
    status_code: Optional[str] = field(default=None)
    content_type: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsAlbTargetGroupTuple:
    kind: ClassVar[str] = "aws_alb_target_group_tuple"
    mapping: ClassVar[Dict[str, Bender]] = {"target_group_arn": S("TargetGroupArn"), "weight": S("Weight")}
    target_group_arn: Optional[str] = field(default=None)
    weight: Optional[int] = field(default=None)


@define(eq=False, slots=False)
class AwsAlbTargetGroupStickinessConfig:
    kind: ClassVar[str] = "aws_alb_target_group_stickiness_config"
    mapping: ClassVar[Dict[str, Bender]] = {"enabled": S("Enabled"), "duration_seconds": S("DurationSeconds")}
    enabled: Optional[bool] = field(default=None)
    duration_seconds: Optional[int] = field(default=None)


@define(eq=False, slots=False)
class AwsAlbForwardActionConfig:
    kind: ClassVar[str] = "aws_alb_forward_action_config"
    mapping: ClassVar[Dict[str, Bender]] = {
        "target_groups": S("TargetGroups", default=[]) >> ForallBend(AwsAlbTargetGroupTuple.mapping),
        "target_group_stickiness_config": S("TargetGroupStickinessConfig")
        >> Bend(AwsAlbTargetGroupStickinessConfig.mapping),
    }
    target_groups: List[AwsAlbTargetGroupTuple] = field(factory=list)
    target_group_stickiness_config: Optional[AwsAlbTargetGroupStickinessConfig] = field(default=None)


@define(eq=False, slots=False)
class AwsAlbAction:
    kind: ClassVar[str] = "aws_alb_action"
    mapping: ClassVar[Dict[str, Bender]] = {
        "type": S("Type"),
        "target_group_arn": S("TargetGroupArn"),
        "authenticate_oidc_config": S("AuthenticateOidcConfig") >> Bend(AwsAlbAuthenticateOidcActionConfig.mapping),
        "authenticate_cognito_config": S("AuthenticateCognitoConfig")
        >> Bend(AwsAlbAuthenticateCognitoActionConfig.mapping),
        "order": S("Order"),
        "redirect_config": S("RedirectConfig") >> Bend(AwsAlbRedirectActionConfig.mapping),
        "fixed_response_config": S("FixedResponseConfig") >> Bend(AwsAlbFixedResponseActionConfig.mapping),
        "forward_config": S("ForwardConfig") >> Bend(AwsAlbForwardActionConfig.mapping),
    }
    type: Optional[str] = field(default=None)
    target_group_arn: Optional[str] = field(default=None)
    authenticate_oidc_config: Optional[AwsAlbAuthenticateOidcActionConfig] = field(default=None)
    authenticate_cognito_config: Optional[AwsAlbAuthenticateCognitoActionConfig] = field(default=None)
    order: Optional[int] = field(default=None)
    redirect_config: Optional[AwsAlbRedirectActionConfig] = field(default=None)
    fixed_response_config: Optional[AwsAlbFixedResponseActionConfig] = field(default=None)
    forward_config: Optional[AwsAlbForwardActionConfig] = field(default=None)


@define(eq=False, slots=False)
class AwsAlbListener:
    kind: ClassVar[str] = "aws_alb_listener"
    mapping: ClassVar[Dict[str, Bender]] = {
        "listener_arn": S("ListenerArn"),
        "load_balancer_arn": S("LoadBalancerArn"),
        "port": S("Port"),
        "protocol": S("Protocol"),
        "certificates": S("Certificates", default=[]) >> ForallBend(AwsAlbCertificate.mapping),
        "ssl_policy": S("SslPolicy"),
        "default_actions": S("DefaultActions", default=[]) >> ForallBend(AwsAlbAction.mapping),
        "alpn_policy": S("AlpnPolicy", default=[]),
    }
    listener_arn: Optional[str] = field(default=None)
    load_balancer_arn: Optional[str] = field(default=None)
    port: Optional[int] = field(default=None)
    protocol: Optional[str] = field(default=None)
    certificates: List[AwsAlbCertificate] = field(factory=list)
    ssl_policy: Optional[str] = field(default=None)
    default_actions: List[AwsAlbAction] = field(factory=list)
    alpn_policy: List[str] = field(factory=list)


@define(eq=False, slots=False)
class AwsAlb(ElbV2Taggable, AwsResource, BaseLoadBalancer):
    kind: ClassVar[str] = "aws_alb"
    api_spec: ClassVar[AwsApiSpec] = AwsApiSpec("elbv2", "describe-load-balancers", "LoadBalancers")
    reference_kinds: ClassVar[ModelReference] = {
        "predecessors": {
            "default": ["aws_vpc", "aws_ec2_subnet", "aws_ec2_security_group"],
            "delete": ["aws_vpc", "aws_ec2_subnet"],
        }
    }
    mapping: ClassVar[Dict[str, Bender]] = {
        "id": S("LoadBalancerName"),
        "name": S("LoadBalancerName"),
        "ctime": S("CreatedTime"),
        "arn": S("LoadBalancerArn"),
        "public_ip_address": S("AvailabilityZones")[0] >> S("LoadBalancerAddresses")[0] >> S("IpAddress"),
        "alb_dns_name": S("DNSName"),
        "alb_canonical_hosted_zone_id": S("CanonicalHostedZoneId"),
        "alb_scheme": S("Scheme"),
        "alb_state": S("State") >> Bend(AwsAlbLoadBalancerState.mapping),
        "alb_type": S("Type"),
        "lb_type": K("alb"),
        "alb_availability_zones": S("AvailabilityZones", default=[]) >> ForallBend(AwsAlbAvailabilityZone.mapping),
        "alb_security_groups": S("SecurityGroups", default=[]),
        "alb_ip_address_type": S("IpAddressType"),
        "alb_customer_owned_ipv4_pool": S("CustomerOwnedIpv4Pool"),
    }
    alb_dns_name: Optional[str] = field(default=None)
    alb_canonical_hosted_zone_id: Optional[str] = field(default=None)
    alb_scheme: Optional[str] = field(default=None)
    alb_state: Optional[AwsAlbLoadBalancerState] = field(default=None)
    alb_availability_zones: List[AwsAlbAvailabilityZone] = field(factory=list)
    alb_security_groups: List[str] = field(factory=list)
    alb_ip_address_type: Optional[str] = field(default=None)
    alb_customer_owned_ipv4_pool: Optional[str] = field(default=None)
    alb_type: Optional[str] = field(default=None)
    alb_listener: List[AwsAlbListener] = field(factory=list)

    @classmethod
    def collect(cls: Type[AwsResource], json: List[Json], builder: GraphBuilder) -> None:
        for js in json:
            lb = AwsAlb.from_api(js)
            tags = builder.client.list("elbv2", "describe-tags", "TagDescriptions", ResourceArns=[lb.arn])
            if tags:
                lb.tags = bend(S("Tags", default=[]) >> ToDict(), tags[0])
            for listener in builder.client.list("elbv2", "describe-listeners", "Listeners", LoadBalancerArn=lb.arn):
                mapped = bend(AwsAlbListener.mapping, listener)
                lb.alb_listener.append(from_json(mapped, AwsAlbListener))
            builder.add_node(lb, js)

    def connect_in_graph(self, builder: GraphBuilder, source: Json) -> None:
        if vpc_id := source.get("VpcId"):
            builder.dependant_node(self, reverse=True, delete_same_as_default=True, clazz=AwsEc2Vpc, id=vpc_id)
        for sg in self.alb_security_groups:
            builder.add_edge(self, EdgeType.default, reverse=True, clazz=AwsEc2SecurityGroup, id=sg)
        for sn in self.alb_availability_zones:
            builder.dependant_node(self, reverse=True, delete_same_as_default=True, clazz=AwsEc2Subnet, id=sn.subnet_id)

    def delete_resource(self, client: AwsClient) -> bool:
        client.call(
            service=self.api_spec.service, action="delete_load_balancer", result_name=None, LoadBalancerArn=self.arn
        )
        return True


@define(eq=False, slots=False)
class AwsAlbMatcher:
    kind: ClassVar[str] = "aws_alb_matcher"
    mapping: ClassVar[Dict[str, Bender]] = {"http_code": S("HttpCode"), "grpc_code": S("GrpcCode")}
    http_code: Optional[str] = field(default=None)
    grpc_code: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsAlbTargetDescription:
    kind: ClassVar[str] = "aws_alb_target_description"
    mapping: ClassVar[Dict[str, Bender]] = {
        "id": S("Id"),
        "port": S("Port"),
        "availability_zone": S("AvailabilityZone"),
    }
    id: Optional[str] = field(default=None)
    port: Optional[int] = field(default=None)
    availability_zone: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsAlbTargetHealth:
    kind: ClassVar[str] = "aws_alb_target_health"
    mapping: ClassVar[Dict[str, Bender]] = {"state": S("State"), "reason": S("Reason"), "description": S("Description")}
    state: Optional[str] = field(default=None)
    reason: Optional[str] = field(default=None)
    description: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsAlbTargetHealthDescription:
    kind: ClassVar[str] = "aws_alb_target_health_description"
    api_spec: ClassVar[AwsApiSpec] = AwsApiSpec("elbv2", "describe-target-health", "TargetHealthDescriptions")
    mapping: ClassVar[Dict[str, Bender]] = {
        "target": S("Target") >> Bend(AwsAlbTargetDescription.mapping),
        "health_check_port": S("HealthCheckPort"),
        "target_health": S("TargetHealth") >> Bend(AwsAlbTargetHealth.mapping),
    }
    target: Optional[AwsAlbTargetDescription] = field(default=None)
    health_check_port: Optional[str] = field(default=None)
    target_health: Optional[AwsAlbTargetHealth] = field(default=None)


@define(eq=False, slots=False)
class AwsAlbTargetGroup(ElbV2Taggable, AwsResource):
    kind: ClassVar[str] = "aws_alb_target_group"
    api_spec: ClassVar[AwsApiSpec] = AwsApiSpec("elbv2", "describe-target-groups", "TargetGroups")
    reference_kinds: ClassVar[ModelReference] = {
        "predecessors": {"default": ["aws_vpc", "aws_alb"], "delete": ["aws_ec2_instance", "aws_vpc"]},
        "successors": {"delete": ["aws_alb"], "default": ["aws_ec2_instance"]},
    }
    mapping: ClassVar[Dict[str, Bender]] = {
        "id": S("TargetGroupName"),
        "tags": S("Tags", default=[]) >> ToDict(),
        "name": S("TargetGroupName"),
        "arn": S("TargetGroupArn"),
        "protocol": S("Protocol"),
        "port": S("Port"),
        "alb_health_check_protocol": S("HealthCheckProtocol"),
        "alb_health_check_port": S("HealthCheckPort"),
        "alb_health_check_enabled": S("HealthCheckEnabled"),
        "alb_health_check_interval_seconds": S("HealthCheckIntervalSeconds"),
        "alb_health_check_timeout_seconds": S("HealthCheckTimeoutSeconds"),
        "alb_healthy_threshold_count": S("HealthyThresholdCount"),
        "alb_unhealthy_threshold_count": S("UnhealthyThresholdCount"),
        "alb_health_check_path": S("HealthCheckPath"),
        "alb_matcher": S("Matcher") >> Bend(AwsAlbMatcher.mapping),
        "target_type": S("TargetType"),
        "alb_protocol_version": S("ProtocolVersion"),
        "alb_ip_address_type": S("IpAddressType"),
    }
    target_type: Optional[str] = field(default=None)
    protocol: Optional[str] = field(default=None)
    port: Optional[int] = field(default=None)
    alb_health_check_protocol: Optional[str] = field(default=None)
    alb_health_check_port: Optional[str] = field(default=None)
    alb_health_check_enabled: Optional[bool] = field(default=None)
    alb_health_check_interval_seconds: Optional[int] = field(default=None)
    alb_health_check_timeout_seconds: Optional[int] = field(default=None)
    alb_healthy_threshold_count: Optional[int] = field(default=None)
    alb_unhealthy_threshold_count: Optional[int] = field(default=None)
    alb_health_check_path: Optional[str] = field(default=None)
    alb_matcher: Optional[AwsAlbMatcher] = field(default=None)
    alb_protocol_version: Optional[str] = field(default=None)
    alb_ip_address_type: Optional[str] = field(default=None)
    alb_target_health: List[AwsAlbTargetHealthDescription] = field(factory=list)

    @classmethod
    def collect(cls: Type[AwsResource], json: List[Json], builder: GraphBuilder) -> None:
        for js in json:
            tg = AwsAlbTargetGroup.from_api(js)
            tags = builder.client.list("elbv2", "describe-tags", "TagDescriptions", ResourceArns=[tg.arn])
            if tags:
                tg.tags = bend(S("Tags", default=[]) >> ToDict(), tags[0])
            for health in builder.client.list(
                "elbv2", "describe-target-health", "TargetHealthDescriptions", TargetGroupArn=tg.arn
            ):
                mapped = bend(AwsAlbTargetHealthDescription.mapping, health)
                tg.alb_target_health.append(from_json(mapped, AwsAlbTargetHealthDescription))
            builder.add_node(tg, js)

    def connect_in_graph(self, builder: GraphBuilder, source: Json) -> None:
        if vpc_id := source.get("VpcId"):
            builder.dependant_node(self, reverse=True, delete_same_as_default=True, clazz=AwsEc2Vpc, id=vpc_id)
        for lb_arn in bend(S("LoadBalancerArns", default=[]), source):
            if lb := builder.node(AwsAlb, arn=lb_arn):
                builder.dependant_node(lb, node=self)
                for th in self.alb_target_health:
                    if th.target and th.target.id:
                        lb.backends.append(th.target.id)
                        builder.dependant_node(self, clazz=AwsEc2Instance, id=th.target.id)

    def delete_resource(self, client: AwsClient) -> bool:
        client.call(
            service=self.api_spec.service, action="delete_target_group", result_name=None, TargetGroupArn=self.arn
        )
        return True


resources: List[Type[AwsResource]] = [AwsAlb, AwsAlbTargetGroup]



from typing import ClassVar, Dict, Optional, Type, List

from attrs import define, field

from resoto_plugin_aws.resource.base import AwsResource, GraphBuilder, AwsApiSpec
from resoto_plugin_aws.resource.ec2 import AwsEc2Subnet, AwsEc2SecurityGroup, AwsEc2Vpc, AwsEc2Instance
from resoto_plugin_aws.utils import ToDict
from resotolib.baseresources import BaseLoadBalancer, EdgeType, BaseAccount, ModelReference  # noqa: F401
from resotolib.json_bender import Bender, S, Bend, bend, ForallBend, K
from resotolib.types import Json
from resoto_plugin_aws.aws_client import AwsClient


# todo: annotate with no serialization annotation
class ElbTaggable:
    def update_resource_tag(self, client: AwsClient, key: str, value: str) -> bool:
        if isinstance(self, AwsResource):
            if spec := self.api_spec:
                client.call(
                    service=spec.service,
                    action="add_tags",
                    result_name=None,
                    LoadBalancerNames=[self.name],
                    Tags=[{"Key": key, "Value": value}],
                )
                return True
            return False
        return False

    def delete_resource_tag(self, client: AwsClient, key: str) -> bool:
        if isinstance(self, AwsResource):
            if spec := self.api_spec:
                client.call(
                    service=spec.service,
                    action="remove_tags",
                    result_name=None,
                    LoadBalancerNames=[self.name],
                    Tags=[{"Key": key}],
                )
                return True
            return False
        return False


@define(eq=False, slots=False)
class AwsElbListener:
    kind: ClassVar[str] = "aws_elb_listener"
    mapping: ClassVar[Dict[str, Bender]] = {
        "protocol": S("Protocol"),
        "load_balancer_port": S("LoadBalancerPort"),
        "instance_protocol": S("InstanceProtocol"),
        "instance_port": S("InstancePort"),
        "ssl_certificate_id": S("SSLCertificateId"),
    }
    protocol: Optional[str] = field(default=None)
    load_balancer_port: Optional[int] = field(default=None)
    instance_protocol: Optional[str] = field(default=None)
    instance_port: Optional[int] = field(default=None)
    ssl_certificate_id: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsElbListenerDescription:
    kind: ClassVar[str] = "aws_elb_listener_description"
    mapping: ClassVar[Dict[str, Bender]] = {
        "listener": S("Listener") >> Bend(AwsElbListener.mapping),
        "policy_names": S("PolicyNames", default=[]),
    }
    listener: Optional[AwsElbListener] = field(default=None)
    policy_names: List[str] = field(factory=list)


@define(eq=False, slots=False)
class AwsElbAppCookieStickinessPolicy:
    kind: ClassVar[str] = "aws_elb_app_cookie_stickiness_policy"
    mapping: ClassVar[Dict[str, Bender]] = {"policy_name": S("PolicyName"), "cookie_name": S("CookieName")}
    policy_name: Optional[str] = field(default=None)
    cookie_name: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsElbLBCookieStickinessPolicy:
    kind: ClassVar[str] = "aws_elb_lb_cookie_stickiness_policy"
    mapping: ClassVar[Dict[str, Bender]] = {
        "policy_name": S("PolicyName"),
        "cookie_expiration_period": S("CookieExpirationPeriod"),
    }
    policy_name: Optional[str] = field(default=None)
    cookie_expiration_period: Optional[int] = field(default=None)


@define(eq=False, slots=False)
class AwsElbPolicies:
    kind: ClassVar[str] = "aws_elb_policies"
    mapping: ClassVar[Dict[str, Bender]] = {
        "app_cookie_stickiness_policies": S("AppCookieStickinessPolicies", default=[])
        >> ForallBend(AwsElbAppCookieStickinessPolicy.mapping),
        "lb_cookie_stickiness_policies": S("LBCookieStickinessPolicies", default=[])
        >> ForallBend(AwsElbLBCookieStickinessPolicy.mapping),
        "other_policies": S("OtherPolicies", default=[]),
    }
    app_cookie_stickiness_policies: List[AwsElbAppCookieStickinessPolicy] = field(factory=list)
    lb_cookie_stickiness_policies: List[AwsElbLBCookieStickinessPolicy] = field(factory=list)
    other_policies: List[str] = field(factory=list)


@define(eq=False, slots=False)
class AwsElbBackendServerDescription:
    kind: ClassVar[str] = "aws_elb_backend_server_description"
    mapping: ClassVar[Dict[str, Bender]] = {
        "instance_port": S("InstancePort"),
        "policy_names": S("PolicyNames", default=[]),
    }
    instance_port: Optional[int] = field(default=None)
    policy_names: List[str] = field(factory=list)


@define(eq=False, slots=False)
class AwsElbHealthCheck:
    kind: ClassVar[str] = "aws_elb_health_check"
    mapping: ClassVar[Dict[str, Bender]] = {
        "target": S("Target"),
        "interval": S("Interval"),
        "timeout": S("Timeout"),
        "unhealthy_threshold": S("UnhealthyThreshold"),
        "healthy_threshold": S("HealthyThreshold"),
    }
    target: Optional[str] = field(default=None)
    interval: Optional[int] = field(default=None)
    timeout: Optional[int] = field(default=None)
    unhealthy_threshold: Optional[int] = field(default=None)
    healthy_threshold: Optional[int] = field(default=None)


@define(eq=False, slots=False)
class AwsElbSourceSecurityGroup:
    kind: ClassVar[str] = "aws_elb_source_security_group"
    mapping: ClassVar[Dict[str, Bender]] = {"owner_alias": S("OwnerAlias"), "group_name": S("GroupName")}
    owner_alias: Optional[str] = field(default=None)
    group_name: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsElb(ElbTaggable, AwsResource, BaseLoadBalancer):
    kind: ClassVar[str] = "aws_elb"
    api_spec: ClassVar[AwsApiSpec] = AwsApiSpec("elb", "describe-load-balancers", "LoadBalancerDescriptions")
    reference_kinds: ClassVar[ModelReference] = {
        "predecessors": {
            "default": ["aws_vpc", "aws_ec2_subnet", "aws_ec2_security_group"],
            "delete": ["aws_vpc", "aws_ec2_subnet", "aws_ec2_instance"],
        },
        "successors": {
            "default": ["aws_ec2_instance"],
        },
    }
    mapping: ClassVar[Dict[str, Bender]] = {
        "id": S("DNSName"),
        "name": S("LoadBalancerName"),
        "lb_type": K("elb"),
        "ctime": S("CreatedTime"),
        "backends": S("Instances", default=[]) >> ForallBend(S("InstanceId")),
        "elb_canonical_hosted_zone_name": S("CanonicalHostedZoneName"),
        "elb_canonical_hosted_zone_name_id": S("CanonicalHostedZoneNameID"),
        "elb_listener_descriptions": S("ListenerDescriptions", default=[])
        >> ForallBend(AwsElbListenerDescription.mapping),
        "elb_policies": S("Policies") >> Bend(AwsElbPolicies.mapping),
        "elb_backend_server_descriptions": S("BackendServerDescriptions", default=[])
        >> ForallBend(AwsElbBackendServerDescription.mapping),
        "elb_availability_zones": S("AvailabilityZones", default=[]),
        "elb_health_check": S("HealthCheck") >> Bend(AwsElbHealthCheck.mapping),
        "elb_source_security_group": S("SourceSecurityGroup") >> Bend(AwsElbSourceSecurityGroup.mapping),
        "scheme": S("Scheme"),
    }
    scheme: Optional[str] = field(default=None)
    elb_canonical_hosted_zone_name: Optional[str] = field(default=None)
    elb_canonical_hosted_zone_name_id: Optional[str] = field(default=None)
    elb_listener_descriptions: List[AwsElbListenerDescription] = field(factory=list)
    elb_policies: Optional[AwsElbPolicies] = field(default=None)
    elb_backend_server_descriptions: List[AwsElbBackendServerDescription] = field(factory=list)
    elb_availability_zones: List[str] = field(factory=list)
    elb_health_check: Optional[AwsElbHealthCheck] = field(default=None)
    elb_source_security_group: Optional[AwsElbSourceSecurityGroup] = field(default=None)

    @classmethod
    def collect(cls: Type[AwsResource], json: List[Json], builder: GraphBuilder) -> None:
        def add_tags(elb: AwsElb) -> None:
            tags = builder.client.list("elb", "describe-tags", "TagDescriptions", LoadBalancerNames=[elb.name])
            if tags:
                elb.tags = bend(S("Tags", default=[]) >> ToDict(), tags[0])

        for js in json:
            instance = cls.from_api(js)
            builder.add_node(instance, js)
            builder.submit_work(add_tags, instance)

    def connect_in_graph(self, builder: GraphBuilder, source: Json) -> None:
        super().connect_in_graph(builder, source)
        if vpc_id := source.get("VPCId"):
            builder.dependant_node(self, reverse=True, delete_same_as_default=True, clazz=AwsEc2Vpc, id=vpc_id)
        for subnet_id in source.get("Subnets", []):
            builder.dependant_node(self, reverse=True, delete_same_as_default=True, clazz=AwsEc2Subnet, id=subnet_id)
        for sg_id in source.get("SecurityGroups", []):
            builder.add_edge(self, EdgeType.default, reverse=True, clazz=AwsEc2SecurityGroup, id=sg_id)
        for instance in self.backends:
            builder.dependant_node(self, clazz=AwsEc2Instance, id=instance)

    def delete_resource(self, client: AwsClient) -> bool:
        client.call(
            service=self.api_spec.service, action="delete_load_balancer", result_name=None, LoadBalancerName=self.name
        )
        return True


resources: List[Type[AwsResource]] = [AwsElb]



from typing import ClassVar, Dict, Optional, List, Type, cast

from attrs import define, field

from resoto_plugin_aws.resource.base import AwsResource, AwsApiSpec, GraphBuilder

from resotolib.baseresources import (  # noqa: F401
    BaseCertificate,
    BasePolicy,
    BaseGroup,
    BaseAccount,
    BaseAccessKey,
    BaseUser,
    BaseDNSZone,
    BaseDNSRecord,
    EdgeType,
    BaseDNSRecordSet,
    ModelReference,
)
from resotolib.json_bender import Bender, S, Bend, ForallBend
from resotolib.types import Json


@define(eq=False, slots=False)
class AwsRoute53ZoneConfig:
    kind: ClassVar[str] = "aws_route53_zone_config"
    mapping: ClassVar[Dict[str, Bender]] = {"comment": S("Comment"), "private_zone": S("PrivateZone")}
    comment: Optional[str] = field(default=None)
    private_zone: Optional[bool] = field(default=None)


@define(eq=False, slots=False)
class AwsRoute53LinkedService:
    kind: ClassVar[str] = "aws_route53_linked_service"
    mapping: ClassVar[Dict[str, Bender]] = {"service_principal": S("ServicePrincipal"), "description": S("Description")}
    service_principal: Optional[str] = field(default=None)
    description: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsRoute53Zone(AwsResource, BaseDNSZone):
    kind: ClassVar[str] = "aws_route53_zone"
    api_spec: ClassVar[AwsApiSpec] = AwsApiSpec("route53", "list-hosted-zones", "HostedZones")
    reference_kinds: ClassVar[ModelReference] = {
        "successors": {
            "default": ["aws_route53_resource_record_set"],
            "delete": [],
        }
    }
    mapping: ClassVar[Dict[str, Bender]] = {
        "id": S("Id"),
        "name": S("Name"),
        "zone_caller_reference": S("CallerReference"),
        "zone_config": S("Config") >> Bend(AwsRoute53ZoneConfig.mapping),
        "zone_resource_record_set_count": S("ResourceRecordSetCount"),
        "zone_linked_service": S("LinkedService") >> Bend(AwsRoute53LinkedService.mapping),
    }
    zone_caller_reference: Optional[str] = field(default=None)
    zone_config: Optional[AwsRoute53ZoneConfig] = field(default=None)
    zone_resource_record_set_count: Optional[int] = field(default=None)
    zone_linked_service: Optional[AwsRoute53LinkedService] = field(default=None)

    @classmethod
    def collect(cls: Type[AwsResource], json: List[Json], builder: GraphBuilder) -> None:
        for js in json:
            zone: AwsRoute53Zone = cast(AwsRoute53Zone, cls.from_api(js))
            builder.add_node(zone, js)
            for rs_js in builder.client.list(
                "route53", "list-resource-record-sets", "ResourceRecordSets", HostedZoneId=zone.id
            ):
                record_set = AwsRoute53ResourceRecordSet.from_api(rs_js)
                builder.add_node(record_set, rs_js)
                builder.add_edge(zone, EdgeType.default, node=record_set)
                for data in record_set.record_values:
                    record = AwsRoute53ResourceRecord(
                        id=record_set.id,
                        name=record_set.name,
                        record_type=record_set.record_type,
                        record_ttl=record_set.record_ttl or 0,
                        record_data=data,
                        record_value="foo",
                    )
                    builder.add_node(record, js)
                    builder.add_edge(record_set, EdgeType.default, node=record)
                    builder.add_edge(record_set, EdgeType.delete, node=record)


@define(eq=False, slots=False)
class AwsRoute53GeoLocation:
    kind: ClassVar[str] = "aws_route53_geo_location"
    mapping: ClassVar[Dict[str, Bender]] = {
        "continent_code": S("ContinentCode"),
        "country_code": S("CountryCode"),
        "subdivision_code": S("SubdivisionCode"),
    }
    continent_code: Optional[str] = field(default=None)
    country_code: Optional[str] = field(default=None)
    subdivision_code: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsRoute53AliasTarget:
    kind: ClassVar[str] = "aws_route53_alias_target"
    mapping: ClassVar[Dict[str, Bender]] = {
        "hosted_zone_id": S("HostedZoneId"),
        "dns_name": S("DNSName"),
        "evaluate_target_health": S("EvaluateTargetHealth"),
    }
    hosted_zone_id: Optional[str] = field(default=None)
    dns_name: Optional[str] = field(default=None)
    evaluate_target_health: Optional[bool] = field(default=None)


@define(eq=False, slots=False)
class AwsRoute53CidrRoutingConfig:
    kind: ClassVar[str] = "aws_route53_cidr_routing_config"
    mapping: ClassVar[Dict[str, Bender]] = {"collection_id": S("CollectionId"), "location_name": S("LocationName")}
    collection_id: Optional[str] = field(default=None)
    location_name: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsRoute53ResourceRecord(AwsResource, BaseDNSRecord):
    kind: ClassVar[str] = "aws_route53_resource_record"
    reference_kinds: ClassVar[ModelReference] = {
        "successors": {
            "default": [],
            "delete": [],
        }
    }


@define(eq=False, slots=False)
class AwsRoute53ResourceRecordSet(AwsResource, BaseDNSRecordSet):
    kind: ClassVar[str] = "aws_route53_resource_record_set"
    reference_kinds: ClassVar[ModelReference] = {
        "successors": {
            "default": ["aws_route53_resource_record"],
            "delete": ["aws_route53_resource_record"],
        }
    }
    mapping: ClassVar[Dict[str, Bender]] = {
        "id": S("Name"),
        "name": S("Name"),
        "record_set_identifier": S("SetIdentifier"),
        "record_type": S("Type"),
        "record_weight": S("Weight"),
        "record_region": S("Region"),
        "record_geo_location": S("GeoLocation") >> Bend(AwsRoute53GeoLocation.mapping),
        "record_fail_over": S("Failover"),
        "record_multi_value_answer": S("MultiValueAnswer"),
        "record_ttl": S("TTL"),
        "record_values": S("ResourceRecords", default=[]) >> ForallBend(S("Value")),
        "record_alias_target": S("AliasTarget") >> Bend(AwsRoute53AliasTarget.mapping),
        "record_health_check_id": S("HealthCheckId"),
        "record_traffic_policy_instance_id": S("TrafficPolicyInstanceId"),
        "record_cidr_routing_config": S("CidrRoutingConfig") >> Bend(AwsRoute53CidrRoutingConfig.mapping),
    }
    record_name: Optional[str] = field(default=None)
    record_set_identifier: Optional[str] = field(default=None)
    record_weight: Optional[int] = field(default=None)
    record_region: Optional[str] = field(default=None)
    record_geo_location: Optional[AwsRoute53GeoLocation] = field(default=None)
    record_fail_over: Optional[str] = field(default=None)
    record_multi_value_answer: Optional[bool] = field(default=None)
    record_alias_target: Optional[AwsRoute53AliasTarget] = field(default=None)
    record_health_check_id: Optional[str] = field(default=None)
    record_traffic_policy_instance_id: Optional[str] = field(default=None)
    record_cidr_routing_config: Optional[AwsRoute53CidrRoutingConfig] = field(default=None)


resources: List[Type[AwsResource]] = [AwsRoute53Zone, AwsRoute53ResourceRecord, AwsRoute53ResourceRecordSet]



from resoto_plugin_aws.resource.base import GraphBuilder
from resoto_plugin_aws.resource.ec2 import AwsEc2InstanceType, AwsEc2Vpc
from resoto_plugin_aws.resource.elbv2 import AwsAlb
from resoto_plugin_aws.resource.iam import AwsIamServerCertificate
from test.resources import round_trip_for
from types import SimpleNamespace
from typing import cast, Any
from resoto_plugin_aws.aws_client import AwsClient
from resoto_plugin_aws.resource.service_quotas import AwsServiceQuota


def test_service_quotas() -> None:
    first, builder = round_trip_for(AwsServiceQuota, "usage", "quota_type")
    assert len(builder.resources_of(AwsServiceQuota)) >= 14


def test_instance_type_quotas() -> None:
    _, builder = round_trip_for(AwsServiceQuota, "usage", "quota_type")
    AwsEc2InstanceType.collect_resources(builder)
    for _, it in builder.global_instance_types.items():
        builder.add_node(it, {})
    expect_quotas(builder, 3)


def test_volume_type_quotas() -> None:
    _, builder = round_trip_for(AwsServiceQuota, "usage", "quota_type")
    for vt in ["gp2", "gp3", "standard", "io1", "io2", "sc1", "st1"]:
        builder.add_node(builder.volume_type(vt), {})  # type: ignore
    expect_quotas(builder, 7)


def test_vpc_quotas() -> None:
    _, builder = round_trip_for(AwsServiceQuota, "usage", "quota_type")
    AwsEc2Vpc.collect_resources(builder)
    expect_quotas(builder, 3)


def test_alb_quotas() -> None:
    _, builder = round_trip_for(AwsServiceQuota, "usage", "quota_type")
    AwsAlb.collect_resources(builder)
    expect_quotas(builder, 2)


def test_iam_server_certificate_quotas() -> None:
    _, builder = round_trip_for(AwsServiceQuota, "usage", "quota_type")
    AwsIamServerCertificate.collect_resources(builder)
    expect_quotas(builder, 2)


def expect_quotas(builder: GraphBuilder, quotas: int) -> None:
    for node, data in builder.graph.nodes(data=True):
        if isinstance(node, AwsServiceQuota):
            node.connect_in_graph(builder, data.get("source", {}))
    # make sure edges have been created
    assert builder.graph.number_of_edges() == quotas


def test_tagging() -> None:
    quota, _ = round_trip_for(AwsServiceQuota, "usage", "quota_type")

    def validate_update_args(**kwargs: Any) -> None:
        assert kwargs["action"] == "tag_resource"
        assert kwargs["ResourceARN"] == quota.arn
        assert kwargs["Tags"] == [{"Key": "foo", "Value": "bar"}]

    def validate_delete_args(**kwargs: Any) -> None:
        assert kwargs["action"] == "untag_resource"
        assert kwargs["ResourceARN"] == quota.arn
        assert kwargs["TagKeys"] == ["foo"]

    client = cast(AwsClient, SimpleNamespace(call=validate_update_args))
    quota.update_resource_tag(client, "foo", "bar")

    client = cast(AwsClient, SimpleNamespace(call=validate_delete_args))
    quota.delete_resource_tag(client, "foo")



import botocore.exceptions
import sys
import inspect
import pathlib
from typing import Iterable
from collections import deque
from itertools import islice
from sqlalchemy import create_engine, func
from sqlalchemy.orm import sessionmaker
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import (
    Column,
    Integer,
    String,
    DateTime,
    PrimaryKeyConstraint,
)
from prompt_toolkit import PromptSession
from prompt_toolkit.history import FileHistory
from prompt_toolkit.completion import WordCompleter
import resotolib.logger
from resotov1.cli import replace_placeholder
from resotolib.utils import split_esc, iec_size_format
from resotolib.args import get_arg_parser, ArgumentParser
from resoto_plugin_aws.utils import aws_session
from resoto_plugin_aws.resources import AWSAccount
from resoto_plugin_aws import current_account_id, AWSPlugin, get_org_accounts

resotolib.logger.getLogger("resoto.cmd").setLevel(resotolib.logger.INFO)
log = resotolib.logger.getLogger("resoto.cmd")

argv = sys.argv[1:]
if "-v" in argv or "--verbose" in argv:
    resotolib.logger.getLogger("resoto.cmd").setLevel(resotolib.logger.DEBUG)
log.info("resoto S3 bucket collector")


def add_args(arg_parser: ArgumentParser) -> None:
    AWSPlugin.add_args(arg_parser)
    arg_parser.add_argument(
        "--aws-s3-db",
        help="Path to local AWS S3 database file",
        dest="aws_s3_db",
        default=":memory:",
    )
    arg_parser.add_argument(
        "--aws-s3-db-debug",
        help="Debug SQL queries (default: False)",
        dest="aws_s3_db_debug",
        action="store_true",
        default=False,
    )
    arg_parser.add_argument(
        "--aws-s3-bucket",
        help="AWS S3 Bucket to collect (default: all)",
        dest="aws_s3_bucket",
        default=None,
    )
    arg_parser.add_argument(
        "--aws-s3-skip-checks",
        help="Skip SQL exists checks (default: False)",
        dest="aws_s3_skip_checks",
        action="store_true",
        default=False,
    )
    arg_parser.add_argument(
        "--aws-s3-collect",
        help="Collect from S3 (default: False)",
        dest="aws_s3_collect",
        action="store_true",
        default=False,
    )
    default_history_file = pathlib.Path.home() / ".resoto_s3_history"
    cli_history_default = None
    if default_history_file.exists():
        cli_history_default = str(default_history_file)
    arg_parser.add_argument(
        "--aws-s3-cli-history",
        help=("Path to AWS S3 CLI history file" " (default: None or ~/.resoto_s3_history if exists)"),
        dest="aws_s3_cli_history",
        type=str,
        default=cli_history_default,
    )


arg_parser = get_arg_parser()
add_args(arg_parser)
arg_parser.parse_args()
engine_args = ""
engine = create_engine(
    f"sqlite:///{ArgumentParser.args.aws_s3_db}",
    echo=ArgumentParser.args.aws_s3_db_debug,
)
Base = declarative_base(bind=engine)
Session = sessionmaker()


class Bucket(Base):
    __tablename__ = "buckets"
    __table_args__ = (PrimaryKeyConstraint("account", "name"),)

    account = Column(String, index=True)
    name = Column(String, index=True)
    ctime = Column(DateTime, index=True)

    def __repr__(self):
        return "<Bucket(account='%s', name='%s', ctime='%s')>" % (
            self.account,
            self.name,
            self.ctime,
        )


class BucketObject(Base):
    __tablename__ = "bucketobjects"
    __table_args__ = (PrimaryKeyConstraint("account", "bucket_name", "name"),)

    account = Column(String, index=True)
    bucket_name = Column(String, index=True)
    name = Column(String, index=True)
    size = Column(Integer, index=True)
    mtime = Column(DateTime, index=True)

    def __repr__(self):
        return "<BucketObject(account= '%s', bucket_name='%s', name='%s', size='%s', mtime='%s')>" % (
            self.account,
            self.bucket_name,
            self.name,
            self.size,
            self.mtime,
        )


Base.metadata.create_all()


def main() -> None:
    if ArgumentParser.args.aws_s3_collect:
        collect()
    cli()


def cli():
    session = PromptSession(history=None)
    history = None
    if ArgumentParser.args.aws_s3_cli_history:
        history = FileHistory(ArgumentParser.args.aws_s3_cli_history)
    session = PromptSession(history=history)
    dbs = Session()
    pwd = "/"
    while True:
        completer = WordCompleter(CliHandler(dbs, pwd).valid_commands)
        try:
            cli_input = session.prompt(f"{pwd} > ", completer=completer)
            if cli_input == "":
                continue

            ch = CliHandler(dbs, pwd)
            for item in ch.evaluate_cli_input(cli_input):
                print(item)
            pwd = ch.pwd

        except KeyboardInterrupt:
            pass
        except EOFError:
            CliHandler.quit("Keyboard Shutdown")
        except (RuntimeError, ValueError) as e:
            log.error(e)
        except Exception:
            log.exception("Caught unhandled exception while processing CLI command")


def collect():
    accounts = get_accounts()
    for account in accounts:
        if not ArgumentParser.args.aws_s3_bucket:
            try:
                buckets = collect_buckets(account)
            except Exception:
                log.exception(f"Failed to collect buckets in {account.rtdname}")
            else:
                for bucket in buckets:
                    try:
                        collect_bucket(account, bucket.name)
                    except Exception:
                        log.exception(f"Failed to collect bucket {bucket.name} in {account.rtdname}")
        else:
            try:
                collect_bucket(account, ArgumentParser.args.aws_s3_bucket)
            except Exception:
                log.exception(f"Failed to collect bucket {ArgumentParser.args.aws_s3_bucket} in {account.rtdname}")


def collect_bucket(account: AWSAccount, bucket_name):
    session = aws_session(account.id, account.role)
    s3 = session.resource("s3")
    dbs = Session()

    log.info(f"Collecting all objects in AWS S3 bucket {bucket_name} in {account.rtdname}")
    bucket = s3.Bucket(bucket_name)
    for bucket_object in bucket.objects.all():
        # TODO: this could be highly optimized via batching
        if (
            not ArgumentParser.args.aws_s3_skip_checks
            and dbs.query(BucketObject)
            .filter_by(account=account.id, bucket_name=bucket_name, name=bucket_object.key)
            .scalar()
            is not None
        ):
            continue

        bo = BucketObject(
            account=account.id,
            bucket_name=bucket_name,
            name=bucket_object.key,
            size=bucket_object.size,
            mtime=bucket_object.last_modified,
        )
        dbs.add(bo)
    dbs.commit()


def collect_buckets(account: AWSAccount):
    session = aws_session(account.id, account.role)
    client = session.client("s3")
    dbs = Session()
    response = client.list_buckets()
    buckets = response.get("Buckets", [])

    log.info(f"Collecting all buckets in {account.rtdname}")

    for bucket in buckets:
        bucket_name = bucket.get("Name")
        bucket_ctime = bucket.get("CreationDate")

        if (
            not ArgumentParser.args.aws_s3_skip_checks
            and dbs.query(Bucket).filter_by(account=account.id, name=bucket_name).scalar() is not None
        ):
            continue

        b = Bucket(account=account.id, name=bucket_name, ctime=bucket_ctime)
        log.info(f"Found bucket {bucket_name} in {account.rtdname}")
        dbs.add(b)
    dbs.commit()

    return dbs.query(Bucket).filter_by(account=account.id)


def authenticated() -> bool:
    try:
        _ = current_account_id()
    except botocore.exceptions.NoCredentialsError:
        log.error("No AWS credentials found")
        return False
    except botocore.exceptions.ClientError as e:
        if e.response["Error"]["Code"] == "AuthFailure":
            log.error("AWS was unable to validate the provided access credentials")
        elif e.response["Error"]["Code"] == "InvalidClientTokenId":
            log.error("AWS was unable to validate the provided security token")
        elif e.response["Error"]["Code"] == "ExpiredToken":
            log.error("AWS security token included in the request is expired")
        else:
            raise
        return False
    return True


def get_accounts():
    if not authenticated():
        log.error("Failed to authenticate")
        return []

    if ArgumentParser.args.aws_assume_current and not ArgumentParser.args.aws_dont_scrape_current:
        log.warning(
            "You specified --aws-assume-current but not --aws-dont-scrape-current! "
            "This will result in the same account being scraped twice and is likely not what you want."
        )

    if ArgumentParser.args.aws_role and ArgumentParser.args.aws_scrape_org:
        accounts = [
            AWSAccount(id=aws_account_id, tags={}, role=ArgumentParser.args.aws_role)
            for aws_account_id in get_org_accounts(filter_current_account=not ArgumentParser.args.aws_assume_current)
            if aws_account_id not in ArgumentParser.args.aws_scrape_exclude_account
        ]
        if not ArgumentParser.args.aws_dont_scrape_current:
            accounts.append(AWSAccount(id=current_account_id(), tags={}))
    elif ArgumentParser.args.aws_role and ArgumentParser.args.aws_account:
        accounts = [
            AWSAccount(id=aws_account_id, tags={}, role=ArgumentParser.args.aws_role)
            for aws_account_id in ArgumentParser.args.aws_account
        ]
    else:
        accounts = [AWSAccount(id=current_account_id(), tags={})]

    return accounts


class CliHandler:
    def __init__(self, dbs, pwd) -> None:
        self.dbs = dbs
        self.commands = {}
        self.pwd = pwd
        self.current_directories = []
        self.current_files = []
        for f, m in inspect.getmembers(self, predicate=inspect.ismethod):
            if f.startswith("cmd_"):
                self.commands[f[4:]] = m
        self.valid_commands = sorted(self.commands.keys())

    def evaluate_cli_input(self, cli_input: str) -> Iterable:
        cli_input = replace_placeholder(cli_input)
        for cmd_chain in split_esc(cli_input, ";"):
            cmds = (cmd.strip() for cmd in split_esc(cmd_chain, "|"))
            items = ()
            for cmd in cmds:
                args = ""
                if " " in cmd:
                    cmd, args = cmd.split(" ", 1)
                if cmd in self.commands:
                    items = self.commands[cmd](items, args)
                else:
                    items = (f"Unknown command: {cmd}",)
                    break
            for item in items:
                yield item

    @staticmethod
    def quit(reason=None):
        log.info(f"Shutting down: {reason}")
        sys.exit(0)

    def cmd_quit(self, items: Iterable, args: str) -> Iterable:
        """Usage: quit

        Quit resoto.
        """
        self.quit("Shutdown requested by CLI input")
        return ()

    def cmd_echo(self, items: Iterable, args: str) -> Iterable:
        """Usage: echo [string]

        Echo a string to the console.

        Can be used for testing string substitution.
        E.g. echo @TODAY@
        """
        yield args

    def cmd_help(self, items: Iterable, args: str) -> Iterable:
        """Usage: help <command>

        Show help text for a command.
        """
        extra_doc = ""
        if args == "":
            args = "help"
            valid_commands = "\n    ".join(self.valid_commands)
            placeholder_help = inspect.getdoc(replace_placeholder)
            extra_doc = f"""\n
{placeholder_help}

Valid commands are:
    {valid_commands}

Note that you can pipe commands using the pipe character (|)
and chain multipe commands using the semicolon (;).
            """
        if args in self.commands:
            doc = inspect.getdoc(self.commands[args])
            if doc is None:
                doc = f"Command {args} has no help text."
        else:
            doc = f"Unknown command: {args}"
        doc += extra_doc
        return (doc,)

    def cmd_head(self, items: Iterable, args: str) -> Iterable:
        """Usage: | head <num> |

        Returns the first num lines.
        """
        num = int(args) if len(args) > 0 else 10
        if num < 0:
            num *= -1
        return islice(items, num)

    def cmd_tail(self, items: Iterable, args: str) -> Iterable:
        """Usage: | tail <num> |

        Returns the last num lines.
        """
        num = int(args) if len(args) > 0 else 10
        if num < 0:
            num *= -1
        return deque(items, num)

    def cmd_accounts(self, items: Iterable, args: str) -> Iterable:
        """Usage: accounts |

        Returns all known accounts.
        """
        for result in self.dbs.query(Bucket.account).distinct().all():
            yield result.account

    def cmd_buckets(self, items: Iterable, args: str) -> Iterable:
        """Usage: buckets |

        Returns all known buckets.
        """
        for result in self.dbs.query(Bucket).all():
            yield result.name

    def cmd_ls(self, items: Iterable, args: str) -> Iterable:
        """Usage: ls [-t|-S] |

        List buckets.
            -t order by mtime
            -S order by size
        """
        if args == "-t":
            order_by = "mtime"
        elif args == "-S":
            order_by = "size"
        else:
            order_by = "dname"
        components = self.pwd.split("/", 3)
        if len(components) == 2:
            account = components[1]
            if account == "":
                account = None
            bucket = None
            s3object = None
        elif len(components) == 3:
            account = components[1]
            bucket = components[2]
            s3object = None
        elif len(components) == 4:
            account = components[1]
            bucket = components[2]
            s3object = components[3] + "/"

        yield f"{'type':<8} {'size':>10} {'modification date':>22} {'name':>40}"
        if account is None:
            for result in list_accounts(self.dbs, order_by=order_by):
                yield self.fprint_result("account", result.size, result.mtime, result.dname)
        elif bucket is None:
            for result in list_buckets(self.dbs, account, order_by=order_by):
                yield self.fprint_result("bucket", result.size, result.mtime, result.dname)
        else:
            for result in list_directories(self.dbs, account, bucket, s3object, order_by=order_by):
                if result.dname == "":
                    directory = "."
                else:
                    directory = result.dname
                yield self.fprint_result("dir", result.size, result.mtime, directory)
            for result in list_objects(self.dbs, account, bucket, s3object, order_by=order_by):
                if len(result.dname) > 0:
                    yield self.fprint_result("file", result.size, result.mtime, result.dname)

    @staticmethod
    def fprint_result(obj_type, obj_size, obj_date, obj_name):
        obj_size = iec_size_format(obj_size)
        return f"{obj_type:<8} {obj_size:>10} {str(obj_date):>22} {obj_name:>40}"

    def cmd_cd(self, items: Iterable, args: str) -> Iterable:
        """Usage: cd <directory>

        Change directory
        """
        if args != "/":
            args = args.rstrip("/")
        if args == ".":
            return ()
        elif args == "..":
            self.pwd = "/".join(self.pwd.split("/")[:-1])
            if self.pwd == "":
                self.pwd = "/"
        elif args.startswith("/"):
            self.pwd = args
        else:
            if self.pwd == "/":
                self.pwd = self.pwd + args
            else:
                self.pwd = self.pwd + "/" + args
        return ()

    def cmd_pwd(self, items: Iterable, args: str) -> Iterable:
        """Usage: pwd

        Show current rirectory
        """
        yield self.pwd

    def cmd_sql(self, items: Iterable, args: str) -> Iterable:
        """Usage: sql <SQL QUERY>

        Execute a raw SQL query
        """
        for row in engine.execute(args):
            yield row


def list_accounts(dbs, order_by="dname"):
    results = (
        dbs.query(
            BucketObject.account.label("dname"),
            func.sum(BucketObject.size).label("size"),
            func.max(BucketObject.mtime).label("mtime"),
        )
        .group_by(BucketObject.account)
        .order_by(order_by)
    )
    return results


def list_buckets(dbs, account, order_by="dname"):
    results = (
        dbs.query(
            BucketObject.account,
            BucketObject.bucket_name.label("dname"),
            func.sum(BucketObject.size).label("size"),
            func.max(BucketObject.mtime).label("mtime"),
        )
        .filter_by(account=account)
        .group_by(BucketObject.account, BucketObject.bucket_name)
        .order_by(order_by)
    )
    return results


def list_objects(dbs, account, bucket_name, dirname, limit=10000, order_by="dname"):
    if dirname is None:
        dirname = ""
    dirname_len = len(dirname) + 1
    results = (
        dbs.query(
            BucketObject.account,
            BucketObject.bucket_name,
            func.substr(BucketObject.name, dirname_len).label("dname"),
            BucketObject.size.label("size"),
            BucketObject.mtime.label("mtime"),
        )
        .filter_by(account=account, bucket_name=bucket_name)
        .filter(BucketObject.name.like(dirname + "%"))
        .filter(func.substr(BucketObject.name, dirname_len).notlike("%/%"))
        .order_by(order_by)
        .limit(limit)
    )
    return results


def list_directories(dbs, account, bucket_name, dirname, order_by="dname"):
    if dirname is None:
        dirname = ""
    dirname_len = len(dirname) + 1

    results = (
        dbs.query(
            BucketObject.account,
            BucketObject.bucket_name,
            func.substr(
                BucketObject.name,
                dirname_len,
                func.instr(func.substr(BucketObject.name, dirname_len), "/"),
            ).label("dname"),
            func.sum(BucketObject.size).label("size"),
            func.max(BucketObject.mtime).label("mtime"),
        )
        .filter_by(account=account, bucket_name=bucket_name)
        .filter(BucketObject.name.like(dirname + "%"))
        .group_by(BucketObject.account, BucketObject.bucket_name, "dname")
        .order_by(order_by)
    )
    return results


if __name__ == "__main__":
    main()



from resoto_plugin_aws.resource.eks import AwsEksCluster, AwsEksNodegroup
from test.resources import round_trip_for
from typing import Any, cast
from types import SimpleNamespace
from resoto_plugin_aws.aws_client import AwsClient


def test_eks_nodegroup() -> None:
    first, builder = round_trip_for(AwsEksCluster)
    assert len(builder.resources_of(AwsEksCluster)) == 1
    assert len(builder.resources_of(AwsEksNodegroup)) == 1


def test_cluster_deletion() -> None:
    cluster, _ = round_trip_for(AwsEksCluster)

    def validate_delete_args(**kwargs: Any) -> None:
        assert kwargs["action"] == "delete_cluster"
        assert kwargs["name"] == cluster.name

    client = cast(AwsClient, SimpleNamespace(call=validate_delete_args))
    cluster.delete_resource(client)


def test_nodegroup_deletion() -> None:
    _, builder = round_trip_for(AwsEksCluster)
    nodegroup = builder.resources_of(AwsEksNodegroup)[0]

    def validate_delete_args(**kwargs: Any) -> None:
        assert kwargs["action"] == "delete_nodegroup"
        assert kwargs["nodegroupName"] == nodegroup.name
        assert kwargs["clusterName"] == nodegroup.cluster_name

    client = cast(AwsClient, SimpleNamespace(call=validate_delete_args))
    nodegroup.delete_resource(client)


def test_tagging() -> None:
    cluster, _ = round_trip_for(AwsEksCluster)

    def validate_update_args(**kwargs: Any) -> None:
        assert kwargs["action"] == "tag_resource"
        assert kwargs["resourceArn"] == cluster.arn
        assert kwargs["tags"] == [{"foo": "bar"}]

    def validate_delete_args(**kwargs: Any) -> None:
        assert kwargs["action"] == "untag_resource"
        assert kwargs["resourceArn"] == cluster.arn
        assert kwargs["tagKeys"] == ["foo"]

    client = cast(AwsClient, SimpleNamespace(call=validate_update_args))
    cluster.update_resource_tag(client, "foo", "bar")

    client = cast(AwsClient, SimpleNamespace(call=validate_delete_args))
    cluster.delete_resource_tag(client, "foo")



import re
from datetime import datetime, timedelta
from typing import ClassVar, Dict, List, Optional, Type, Tuple

from attr import define, field

from resoto_plugin_aws.aws_client import AwsClient
from resoto_plugin_aws.resource.base import AwsApiSpec, AwsResource, GraphBuilder
from resotolib.baseresources import BaseAccount, ModelReference  # noqa: F401
from resotolib.json import from_json
from resotolib.json_bender import S, Bend, Bender, ForallBend, bend
from resotolib.types import Json
from resotolib.utils import chunks


# todo: annotate with no serialization annotation
class CloudwatchTaggable:
    def update_resource_tag(self, client: AwsClient, key: str, value: str) -> bool:
        if isinstance(self, AwsResource):
            if spec := self.api_spec:
                client.call(
                    service=spec.service,
                    action="tag_resource",
                    result_name=None,
                    ResourceARN=self.arn,
                    Tags=[{"Key": key, "Value": value}],
                )
                return True
            return False
        return False

    def delete_resource_tag(self, client: AwsClient, key: str) -> bool:
        if isinstance(self, AwsResource):
            if spec := self.api_spec:
                client.call(
                    service=spec.service,
                    action="untag_resource",
                    result_name=None,
                    ResourceARN=self.arn,
                    TagKeys=[key],
                )
                return True
            return False
        return False


@define(eq=False, slots=False)
class AwsCloudwatchDimension:
    kind: ClassVar[str] = "aws_cloudwatch_dimension"
    mapping: ClassVar[Dict[str, Bender]] = {"name": S("Name"), "value": S("Value")}
    name: Optional[str] = field(default=None)
    value: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsCloudwatchMetric:
    kind: ClassVar[str] = "aws_cloudwatch_metric"
    mapping: ClassVar[Dict[str, Bender]] = {
        "namespace": S("Namespace"),
        "metric_name": S("MetricName"),
        "dimensions": S("Dimensions", default=[]) >> ForallBend(AwsCloudwatchDimension.mapping),
    }
    namespace: Optional[str] = field(default=None)
    metric_name: Optional[str] = field(default=None)
    dimensions: List[AwsCloudwatchDimension] = field(factory=list)


@define(eq=False, slots=False)
class AwsCloudwatchMetricStat:
    kind: ClassVar[str] = "aws_cloudwatch_metric_stat"
    mapping: ClassVar[Dict[str, Bender]] = {
        "metric": S("Metric") >> Bend(AwsCloudwatchMetric.mapping),
        "period": S("Period"),
        "stat": S("Stat"),
        "unit": S("Unit"),
    }
    metric: Optional[AwsCloudwatchMetric] = field(default=None)
    period: Optional[int] = field(default=None)
    stat: Optional[str] = field(default=None)
    unit: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsCloudwatchMetricDataQuery:
    kind: ClassVar[str] = "aws_cloudwatch_metric_data_query"
    mapping: ClassVar[Dict[str, Bender]] = {
        "id": S("Id"),
        "metric_stat": S("MetricStat") >> Bend(AwsCloudwatchMetricStat.mapping),
        "expression": S("Expression"),
        "label": S("Label"),
        "return_data": S("ReturnData"),
        "period": S("Period"),
        "account_id": S("AccountId"),
    }
    id: Optional[str] = field(default=None)
    metric_stat: Optional[AwsCloudwatchMetricStat] = field(default=None)
    expression: Optional[str] = field(default=None)
    label: Optional[str] = field(default=None)
    return_data: Optional[bool] = field(default=None)
    period: Optional[int] = field(default=None)
    account_id: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsCloudwatchAlarm(CloudwatchTaggable, AwsResource):
    kind: ClassVar[str] = "aws_cloudwatch_alarm"
    api_spec: ClassVar[AwsApiSpec] = AwsApiSpec("cloudwatch", "describe-alarms", "MetricAlarms")
    reference_kinds: ClassVar[ModelReference] = {
        "predecessors": {"default": ["aws_ec2_instance"], "delete": ["aws_ec2_instance"]},
    }
    mapping: ClassVar[Dict[str, Bender]] = {
        "id": S("AlarmName"),
        "name": S("AlarmName"),
        "mtime": S("AlarmConfigurationUpdatedTimestamp"),
        "arn": S("AlarmArn"),
        "cloudwatch_alarm_description": S("AlarmDescription"),
        "cloudwatch_actions_enabled": S("ActionsEnabled"),
        "cloudwatch_ok_actions": S("OKActions", default=[]),
        "cloudwatch_alarm_actions": S("AlarmActions", default=[]),
        "cloudwatch_insufficient_data_actions": S("InsufficientDataActions", default=[]),
        "cloudwatch_state_value": S("StateValue"),
        "cloudwatch_state_reason": S("StateReason"),
        "cloudwatch_state_reason_data": S("StateReasonData"),
        "cloudwatch_state_updated_timestamp": S("StateUpdatedTimestamp"),
        "cloudwatch_metric_name": S("MetricName"),
        "cloudwatch_namespace": S("Namespace"),
        "cloudwatch_statistic": S("Statistic"),
        "cloudwatch_extended_statistic": S("ExtendedStatistic"),
        "cloudwatch_dimensions": S("Dimensions", default=[]) >> ForallBend(AwsCloudwatchDimension.mapping),
        "cloudwatch_period": S("Period"),
        "cloudwatch_unit": S("Unit"),
        "cloudwatch_evaluation_periods": S("EvaluationPeriods"),
        "cloudwatch_datapoints_to_alarm": S("DatapointsToAlarm"),
        "cloudwatch_threshold": S("Threshold"),
        "cloudwatch_comparison_operator": S("ComparisonOperator"),
        "cloudwatch_treat_missing_data": S("TreatMissingData"),
        "cloudwatch_evaluate_low_sample_count_percentile": S("EvaluateLowSampleCountPercentile"),
        "cloudwatch_metrics": S("Metrics", default=[]) >> ForallBend(AwsCloudwatchMetricDataQuery.mapping),
        "cloudwatch_threshold_metric_id": S("ThresholdMetricId"),
    }
    arn: Optional[str] = field(default=None)
    cloudwatch_alarm_description: Optional[str] = field(default=None)
    cloudwatch_actions_enabled: Optional[bool] = field(default=None)
    cloudwatch_ok_actions: List[str] = field(factory=list)
    cloudwatch_alarm_actions: List[str] = field(factory=list)
    cloudwatch_insufficient_data_actions: List[str] = field(factory=list)
    cloudwatch_state_value: Optional[str] = field(default=None)
    cloudwatch_state_reason: Optional[str] = field(default=None)
    cloudwatch_state_reason_data: Optional[str] = field(default=None)
    cloudwatch_state_updated_timestamp: Optional[datetime] = field(default=None)
    cloudwatch_metric_name: Optional[str] = field(default=None)
    cloudwatch_namespace: Optional[str] = field(default=None)
    cloudwatch_statistic: Optional[str] = field(default=None)
    cloudwatch_extended_statistic: Optional[str] = field(default=None)
    cloudwatch_dimensions: List[AwsCloudwatchDimension] = field(factory=list)
    cloudwatch_period: Optional[int] = field(default=None)
    cloudwatch_unit: Optional[str] = field(default=None)
    cloudwatch_evaluation_periods: Optional[int] = field(default=None)
    cloudwatch_datapoints_to_alarm: Optional[int] = field(default=None)
    cloudwatch_threshold: Optional[float] = field(default=None)
    cloudwatch_comparison_operator: Optional[str] = field(default=None)
    cloudwatch_treat_missing_data: Optional[str] = field(default=None)
    cloudwatch_evaluate_low_sample_count_percentile: Optional[str] = field(default=None)
    cloudwatch_metrics: List[AwsCloudwatchMetricDataQuery] = field(factory=list)
    cloudwatch_threshold_metric_id: Optional[str] = field(default=None)

    def connect_in_graph(self, builder: GraphBuilder, source: Json) -> None:
        super().connect_in_graph(builder, source)
        for dimension in self.cloudwatch_dimensions:
            builder.dependant_node(
                self, reverse=True, delete_same_as_default=True, kind="aws_ec2_instance", id=dimension.value
            )

    def delete_resource(self, client: AwsClient) -> bool:
        client.call(service=self.api_spec.service, action="delete_alarms", result_name=None, AlarmNames=[self.name])
        return True


@define(hash=True, frozen=True)
class AwsCloudwatchQuery:
    metric_name: str
    namespace: str
    dimensions: Tuple[Tuple[str, str], ...]
    period: timedelta
    ref_id: str
    metric_id: str

    def to_json(self) -> Json:
        return {
            "Id": self.metric_id,
            "MetricStat": {
                "Metric": {
                    "Namespace": self.namespace,
                    "MetricName": self.metric_name,
                    "Dimensions": [{"Name": k, "Value": v} for k, v in self.dimensions],
                },
                "Period": int((self.period.total_seconds() / 60) * 60),  # round to the next 60 seconds
                "Stat": "Sum",
                "Unit": "Count",
            },
            "ReturnData": True,
        }

    @staticmethod
    def create(
        metric_name: str,
        namespace: str,
        period: timedelta,
        ref_id: str,
        metric_id: Optional[str] = None,
        **dimensions: str,
    ) -> "AwsCloudwatchQuery":
        dims = "_".join(f"{k}+{v}" for k, v in dimensions.items())
        rid = metric_id or re.sub("\\W", "_", f"{metric_name}-{namespace}-{dims}".lower())
        return AwsCloudwatchQuery(
            metric_name=metric_name,
            namespace=namespace,
            period=period,
            dimensions=tuple(dimensions.items()),
            ref_id=ref_id,
            metric_id=rid,
        )


@define(eq=False, slots=False)
class AwsCloudwatchMessageData:
    mapping: ClassVar[Dict[str, Bender]] = {"code": S("Code"), "value": S("Value")}
    code: Optional[str] = field(default=None)
    value: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsCloudwatchMetricData:
    mapping: ClassVar[Dict[str, Bender]] = {
        "id": S("Id"),
        "label": S("Label"),
        "metric_timestamps": S("Timestamps", default=[]),
        "metric_values": S("Values", default=[]),
        "metric_status_code": S("StatusCode"),
        "metric_messages": S("Messages", default=[]) >> ForallBend(AwsCloudwatchMessageData.mapping),
    }
    id: str = field(default=None)
    label: Optional[str] = field(default=None)
    metric_timestamps: List[datetime] = field(factory=list)
    metric_values: List[float] = field(factory=list)
    metric_status_code: Optional[str] = field(default=None)
    metric_messages: List[AwsCloudwatchMessageData] = field(factory=list)

    def first_non_zero(self) -> Optional[Tuple[datetime, float]]:
        for timestamp, value in zip(self.metric_timestamps, self.metric_values):
            if value != 0:
                return timestamp, value
        return None

    @staticmethod
    def query_for(
        client: AwsClient,
        queries: List[AwsCloudwatchQuery],
        start_time: datetime,
        end_time: datetime,
        scan_desc: bool = True,
    ) -> "Dict[AwsCloudwatchQuery, AwsCloudwatchMetricData]":
        lookup = {q.metric_id: q for q in queries}
        result: Dict[AwsCloudwatchQuery, AwsCloudwatchMetricData] = {}
        # the api only allows for up to 500 metrics at once
        for chunk in chunks(queries, 499):
            part = client.list(
                "cloudwatch",
                "get-metric-data",
                "MetricDataResults",
                MetricDataQueries=[a.to_json() for a in chunk],
                StartTime=start_time,
                EndTime=end_time,
                ScanBy="TimestampDescending" if scan_desc else "TimestampAscending",
            )
            for single in part:
                metric = from_json(bend(AwsCloudwatchMetricData.mapping, single), AwsCloudwatchMetricData)
                result[lookup[metric.id]] = metric

        return result


resources: List[Type[AwsResource]] = [AwsCloudwatchAlarm]
